#!/usr/bin/env python
"""
Utility functions for dfvue.

The utility functions do not depend on the dfvue class.
Functions depending on the class are in dfvmethods.

This module was written by Matthias Cuntz while at Institut National de
Recherche pour l'Agriculture, l'Alimentation et l'Environnement (INRAE), Nancy,
France.

:copyright: Copyright 2023- Matthias Cuntz - mc (at) macu (dot) de
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following functions are provided:

.. autosummary::
   clone_dfvmain
   format_coord_scatter
   list_intersection
   parse_entry
   vardim2var

History
   * Written Jul 2023 by Matthias Cuntz (mc (at) macu (dot) de)
   * Use dfvMain directly for cloning window, Jun 2014, Matthias Cuntz
   * Use CustomTkinter, Jun 2024, Matthias Cuntz
   * Use mix of grid and pack layout manager, Jun 2024, Matthias Cuntz
   * Use CustomTkinter only if installed, Jun 2024, Matthias Cuntz
   * Allow list in window title in clone_dfvmain, Oct 2024, Matthias Cuntz
   * Remove [ms] from check for datetime in format_coord on axes2,
     Oct 2024, Matthias Cuntz
   * Back to pack layout manager for resizing, Nov 2024, Matthias Cuntz
   * Increased digits in format_coord_scatter, Jan 2025, Matthias Cuntz
   * Add parse_entry from dfvreadcsv, Jan 2025, Matthias Cuntz

"""
import tkinter as tk
try:
    from customtkinter import CTkToplevel as Toplevel
except ModuleNotFoundError:
    from tkinter import Toplevel
from math import isfinite
import numpy as np
import matplotlib.dates as mpld
import dfvue


__all__ = ['clone_dfvmain',
           'format_coord_scatter',
           'list_intersection',
           'vardim2var']


#
# Clone the main window
#

def clone_dfvmain(widget):
    """
    Duplicate the main dfvue window.

    Parameters
    ----------
    widget : dfvue.dfvMain
        widget of dfvMain class.

    Returns
    -------
    Another dfvue window will be created.

    Examples
    --------
    >>> self.newwin = ctk.CTkButton(
    ...     self.rowwin, text="New Window",
    ...     command=partial(clone_dfvmain, self.master))

    """
    if widget.name != 'dfvMain':
        print('clone_dfvmain failed. Widget should be dfvMain.')
        print('widget.name is: ', widget.name)
        import sys
        sys.exit()

    root = Toplevel()
    root.name = 'dfvClone'
    root.top = widget.top
    if root.top.csvfile:
        tit = f"Secondary dfvue {root.top.csvfile}"
    else:
        tit = "Secondary dfvue"
    root.title(tit)
    root.geometry('1000x800+150+100')

    # https://stackoverflow.com/questions/46505982/is-there-a-way-to-clone-a-tkinter-widget
    cls = widget.__class__
    clone = cls(root)
    try:
        for key in widget.configure():
            if key != 'class':
                clone.configure({key: widget.cget(key)})
    except TypeError:
        cls = dfvue.dfvMain
        clone = cls(root)
        clone.pack(fill=tk.BOTH, expand=1)

    return clone


#
# How to write the value of the data point below the pointer
#

def format_coord_scatter(x, y, ax, ax2, xdtype, ydtype, y2dtype):
    """
    Formatter function for scatter plot with left and right axis
    having the same x-axis.

    Parameters
    ----------
    x, y : float
        Data coordinates of `ax2`.
    ax, ax2: matplotlib.axes._subplots.AxesSubplot
        Matplotlib axes object for left-hand and right-hand y-axis, resp.
    xdtype, ydtype, y2dtype: numpy.dtype
        Numpy dtype of data of x-values (xdtype), left-hand side y-values
        (ydtype), and right-hand side y-values (y2dtype)

    Returns
    -------
    String with left-hand side and right hand-side coordinates.

    Examples
    --------
    >>> ax = plt.subplot(111)
    >>> ax2 = ax.twinx()
    >>> ax.plot(xx, yy)
    >>> ax2.plot(xx, yy2)
    >>> ax2.format_coord = lambda x, y: format_coord_scatter(
    ...     x, y, ax, ax2, xx.dtype, yy.dtype, yy2.dtype)

    """
    # convert to display coords
    # https://stackoverflow.com/questions/21583965/matplotlib-cursor-value-with-two-axes
    display_coord = ax2.transData.transform((x, y))
    # convert back to data coords with respect to ax
    inv      = ax.transData.inverted()
    ax_coord = inv.transform(display_coord)

    # Special treatment for datetime
    # https://stackoverflow.com/questions/49267011/matplotlib-datetime-from-event-coordinates
    if xdtype.type == np.dtype('datetime64').type:
        xstr = mpld.num2date(x).strftime('%Y-%m-%d %H:%M:%S')
    else:
        xstr  = '{:.6g}'.format(x)
    if ydtype.type == np.dtype('datetime64').type:
        ystr = mpld.num2date(ax_coord[1]).strftime('%Y-%m-%d %H:%M:%S')
    else:
        ystr  = '{:.6g}'.format(ax_coord[1])
    if y2dtype.type == np.dtype('datetime64').type:
        y2str = mpld.num2date(y).strftime('%Y-%m-%d %H:%M:%S')
    else:
        y2str = '{:.6g}'.format(y)
    out = f'Left: ({xstr}, {ystr}) Right: ({xstr}, {y2str})'
    return out


#
# Intersection of two lists
#

def list_intersection(lst1, lst2):
    """
    Intersection of two lists.

    From:
    https://stackoverflow.com/questions/3697432/how-to-find-list-intersection
    Using list comprehension for small lists and set() method with builtin
    intersection for longer lists.

    Parameters
    ----------
    lst1, lst2 : list
        Python lists

    Returns
    -------
    list
        List with common elements in both input lists.

    Examples
    --------
    >>> lst1 = [ 4, 9, 1, 17, 11, 26, 28, 28, 26, 66, 91]
    >>> lst2 = [9, 9, 74, 21, 45, 11, 63]
    >>> print(Intersection(lst1, lst2))
    [9, 11]

    """
    if (len(lst1) > 10) or (len(lst2) > 10):
        return list(set(lst1).intersection(lst2))
    else:
        return [ ll for ll in lst1 if ll in lst2 ]


def parse_entry(text):
    """
    Convert text string to correct data type

    Parse an entry field to None, bool, int, float, list, dict

    Parameters
    ----------
    text : str
        String from entry field

    Returns
    -------
    None, bool, int, float, list, dict

    Examples
    --------
    >>> parse_entry('7')
    7
    >>> parse_entry('7,3')
    [7, 3]

    """
    if text == 'None':
        # None
        tt = None
    elif text == 'True':
        # bool True
        tt = True
    elif text == 'False':
        # bool False
        tt = False
    elif ':' in text:
        # dict or str
        try:
            tt = eval(f'{{{text}}}')
        except SyntaxError:
            tt = text
    elif ',' in text:
        # list or str
        try:
            tt = eval(f'[{text}]')
        except SyntaxError:
            tt = text
    else:
        try:
            # int
            tt = int(text)
        except ValueError:
            try:
                # float
                tt = float(text)
            except ValueError:
                # str
                tt = text
            try:
                if not isfinite(tt):
                    # keep NaN and Inf string
                    tt = text
            except TypeError:
                pass
    return tt


def vardim2var(vardim):
    """
    Extract variable name from 'variable (dim1=ndim1,)' string.

    Parameters
    ----------
    vardim : string
        Variable name with dimensions, such as 'latitude (lat=32,lon=64)'.

    Returns
    -------
    string
        Variable name.

    Examples
    --------
    >>> vardim2var('latitude (lat=32,lon=64)')
    latitude

    """
    return vardim[0:vardim.rfind('(')].rstrip()



#!/usr/bin/env python
"""
Widget functions.

Convenience functions for adding Tkinter widgets.

This module was written by Matthias Cuntz while at Institut National de
Recherche pour l'Agriculture, l'Alimentation et l'Environnement (INRAE), Nancy,
France.

:copyright: Copyright 2020- Matthias Cuntz - mc (at) macu (dot) de
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following functions are provided:

.. autosummary::
   callurl
   Tooltip
   add_checkbutton
   add_combobox
   add_entry
   add_imagemenu
   add_menu
   add_scale
   add_spinbox
   add_tooltip
   Treeview

History
   * Written Nov-Dec 2020 by Matthias Cuntz (mc (at) macu (dot) de)
   * Added tooltips to all widgets with class Tooltip,
     Jan 2021, Matthias Cuntz
   * Added add_tooltip widget, Jan 2021, Matthias Cuntz
   * add_spinbox returns also label widget, Jan 2021, Matthias Cuntz
   * padlabel for add_entry to add space to previous widget,
     Jul 2023, Matthias Cuntz
   * labelwidth for add_entry to align columns with pack,
     Jul 2023, Matthias Cuntz
   * Replace tk constants with strings such as tk.LEFT with 'left',
     Jul 2023, Matthias Cuntz
   * Use Hovertip from local copy of tooltip.py, Jul 2023, Matthias Cuntz
   * Added Treeview class with optional horizontal and vertical scroolbars,
     Jul 2023, Matthias Cuntz
   * Added callurl function, Dec 2023, Matthias Cuntz
   * Use CustomTkinter, Jun 2024, Matthias Cuntz
   * Use CustomTkinter only if installed, Jun 2024, Matthias Cuntz
   * Small bugfix in Combobox if no CustomTkinter, Nov 2024, Matthias Cuntz
   * Pass width to Checkbutton if CustomTkinter, Dec 2024, Matthias Cuntz
   * Pass padx for space between label and combobox, Dec 2024, Matthias Cuntz
   * Use CustomTkinter also in add_menu and add_scale,
     Dec 2024, Matthias Cuntz
   * Bugfix: did not make new frame in add_spinbox, Dec 2024, Matthias Cuntz

"""
import tkinter as tk
import tkinter.ttk as ttk
try:
    from customtkinter import CTkFrame as Frame
    from customtkinter import CTkLabel as Label
    from customtkinter import CTkCheckBox as Checkbutton
    from customtkinter import CTkComboBox as Combobox
    from customtkinter import CTkEntry as Entry
    from customtkinter import CTkOptionMenu as Menubutton
    from customtkinter import CTkSlider as Scale
    from customtkinter import CTkScrollbar as Scrollbar
    ihavectk = True
except ModuleNotFoundError:
    from tkinter.ttk import Frame
    from tkinter.ttk import Label
    from tkinter.ttk import Checkbutton
    from tkinter.ttk import Combobox
    from tkinter.ttk import Entry
    from tkinter.ttk import Menubutton
    from tkinter.ttk import Scale
    from tkinter.ttk import Scrollbar
    ihavectk = False
import webbrowser

from .tooltip import Hovertip


__all__ = ['callurl', 'Tooltip',
           'add_checkbutton', 'add_combobox', 'add_entry',
           'add_imagemenu', 'add_menu', 'add_scale', 'add_spinbox',
           'add_tooltip',
           'Treeview']


# https://stackoverflow.com/questions/23482748/how-to-create-a-hyperlink-with-a-label-in-tkinter
def callurl(url):
    """
    Open url in external web browser

    Parameters
    ----------
    url : str
        html url

    Returns
    -------
    Opens *url* in external web browser

    Examples
    --------
    >>> opthead = Frame(self)
    >>> opthead.pack(side='top', fill='x')
    >>> optheadlabel1 = Label(opthead, text='Options for')
    >>> optheadlabel1.pack(side='left')
    >>> ttk.Style().configure('blue.TLabel', foreground='blue')
    >>> optheadlabel2 = Label(opthead, text='pandas.read_csv',
    ...                              style='blue.TLabel')
    >>> optheadlabel2.pack(side='left')
    >>> font = tkfont.Font(optheadlabel2, optheadlabel2.cget("font"))
    >>> font.configure(underline=True)
    >>> optheadlabel2.configure(font=font)
    >>> optheadlabel2.bind("<Button-1>",
    ...                    lambda e: callurl("https://pandas.pydata.org/docs/reference/api/pandas.read_csv.html"))

    """
    webbrowser.open_new(url)


class Tooltip(Hovertip):
    """
    A tooltip that pops up when a mouse hovers over an anchor widget.

    This is a copy/extension of the class Hovertip of Python's
    idlelib/tooltip.py.
    In addition, it sets the foreground colour to see the tip also in
    macOS dark mode, and displays a textvariable rather than simple text
    so one can change the tip during run time.

    """
    def __init__(self, anchor_widget, text, hover_delay=1000):
        """
        Create a text tooltip with a mouse hover delay.
        anchor_widget: the widget next to which the tooltip will be shown
        text: tk.StringVar with text to display
        hover_delay: time to delay before showing the tooltip, in milliseconds
        Note that a widget will only be shown when showtip() is called,
        e.g. after hovering over the anchor widget with the mouse for enough
        time.

        """
        super(Tooltip, self).__init__(anchor_widget, text,
                                      hover_delay=hover_delay)

    def showcontents(self):
        # light yellow = #ffffe0
        if ihavectk:
            label = Label(self.tipwindow, textvariable=self.text,
                          fg_color="#ffffe0", text_color="#000000",
                          justify='left', padx=1, pady=1)
        else:
            label = tk.Label(self.tipwindow, textvariable=self.text,
                             background="#ffffe0", foreground="#000000",
                             justify='left', relief='flat', borderwidth=0,
                             padx=1, pady=1)
        label.pack()
        # label.grid()


def add_checkbutton(frame, label="", value=False, command=None, tooltip="",
                    **kwargs):
    """
    Add a left-aligned Checkbutton.

    Parameters
    ----------
    frame : CustomTkinter widget
        Parent widget
    label : str, optional
        Text that appears on the checkbutton (default: "")
    value : bool, optional
        Initial state of the checkbutton (default: False)
    command : function, optional
        Function to be called whenever the state of the
        checkbutton changes (default: None).
    tooltip : str, optional
        Tooltip appearing after one second when hovering over
        the checkbutton (default: "" = no tooltip)
    **kwargs : option=value pairs, optional
        All other options will be passed to Checkbutton

    Returns
    -------
    tk.StringVar, tk.BooleanVar
        variable for the text on the checkbutton,
        control variable tracking current state of checkbutton
    tk.StringVar
        variable for the text of the tooltip, if given.

    Examples
    --------
    >>> self.rowzxy = Frame(self)
    >>> self.rowzxy.pack(side='top', fill='x')
    >>> self.inv_xlbl, self.inv_x = add_checkbutton(
    ...     self.rowzxy, label="invert x", value=False, command=self.checked)

    """
    iframe = Frame(frame)
    check_label = tk.StringVar()
    check_label.set(label)
    bvar = tk.BooleanVar(value=value)
    if ihavectk and ('width' not in kwargs):
        width = len(label) * 9
        kwargs.update({'width': width})
    cb = Checkbutton(iframe, variable=bvar, textvariable=check_label,
                     command=command, **kwargs)
    cb.pack(side='left', padx=3)
    if tooltip:
        ttip = tk.StringVar()
        ttip.set(tooltip)
        cbtip = Tooltip(cb, ttip)
        return iframe, check_label, bvar, ttip
    else:
        return iframe, check_label, bvar


def add_combobox(frame, label="", values=[], command=None, tooltip="",
                 **kwargs):
    """
    Add a left-aligned Combobox with a Label before.

    Parameters
    ----------
    frame : CustomTkinter widget
        Parent widget
    label : str, optional
        Text that appears in front of the combobox (default: "")
    values : list of str, optional
        The choices that will appear in the drop-down menu (default: [])
    command : function, optional
        Handler function to be bound to the combobox for the event
        <<ComboboxSelected>> (default: None).
    tooltip : str, optional
        Tooltip appearing after one second when hovering over
        the combobox (default: "" = no tooltip)
    padx : int, optional
        Extra space in px added left and right of the label text (default: 1)
    **kwargs : option=value pairs, optional
        All other options will be passed to Combobox

    Returns
    -------
    tk.StringVar, Combobox
        variable for the text before the combobox, combobox widget
    tk.StringVar
        variable for the text of the tooltip, if given.

    Examples
    --------
    >>> self.rowzxy = Frame(self)
    >>> self.rowzxy.pack(side='top', fill='x')
    >>> self.xlbl, self.x = add_combobox(
    ...     self.rowzxy, label="x", values=columns, command=self.selected)

    """
    iframe = Frame(frame)
    width = kwargs.pop('width', 25)
    cb_label = tk.StringVar()
    cb_label.set(label)
    lkwargs = {'textvariable': cb_label}
    if ihavectk:
        lkwargs.update({'padx': kwargs.pop('padx', 1)})
    else:
        _ = kwargs.pop('padx', 1)
    label = Label(iframe, **lkwargs)
    label.pack(side='left')
    if ihavectk:
        cb = Combobox(iframe, values=values, width=width, command=command,
                      **kwargs)
    else:
        cb = ttk.Combobox(iframe, values=values, width=width, **kwargs)
        # long = len(max(values, key=len))
        # cb.configure(width=(max(20, long//2)))
        if command is not None:
            cb.bind("<<ComboboxSelected>>", command)
    cb.pack(side='left')
    if tooltip:
        ttip = tk.StringVar()
        ttip.set(tooltip)
        cbtip = Tooltip(cb, ttip)
        return iframe, cb_label, cb, ttip
    else:
        return iframe, cb_label, cb


def add_entry(frame, label="", text="", command=None, tooltip="",
              padlabel=0, labelwidth=None, **kwargs):
    """
    Add a left-aligned Entry with a Label before.

    Parameters
    ----------
    frame : CustomTkinter widget
        Parent widget
    label : str, optional
        Text that appears in front of the entry (default: "")
    text : str, optional
        Initial text in the entry area (default: "")
    command : function or list of functions, optional
        Handler function to be bound to the entry for the events
        '<FocusOut>', <Return>, '<Key-Return>', and <KP_Enter> (default: None).
        If list is given than command[0] is bound to '<FocusOut>' and
        command[1] is bound to the 3 events
    tooltip : str, optional
        Tooltip appearing after one second when hovering over
        the entry (default: "" = no tooltip)
    padlabel : int, optional
        Prepend number of spaces to create distance to other widgets (default: 0)
    padx : int, optional
        Extra space in px added left and right of the label text (default: 1)
    labelwidth : int, optional
        If given, set width of Label
    **kwargs : option=value pairs, optional
        All other options will be passed to Entry

    Returns
    -------
    tk.StringVar, tk.StringVar
        variable for the text before the entry,
        variable for the text in the entry area
    tk.StringVar
        variable for the text of the tooltip, if given.

    Examples
    --------
    >>> self.rowxyopt = Frame(self)
    >>> self.rowxyopt.pack(side='top', fill='x')
    >>> self.lslbl, self.ls = add_entry(
    ...     self.rowxyopt, label="ls", text='-',
    ...     width=4, command=self.selected_y)

    """
    # label
    iframe = Frame(frame)
    entry_label = tk.StringVar()
    nlab = len(label) + padlabel
    lab = f'{label:>{nlab}s}'
    entry_label.set(lab)
    lkwargs = {'textvariable': entry_label}
    if labelwidth is not None:
        lkwargs.update({'width': labelwidth})
    # if labelwidth is None:
    #     labelwidth = len(lab)
    #     if ihavectk:
    #         labelwidth *= 9
    # lkwargs.update({'width': labelwidth})
    if ihavectk:
        lkwargs.update({'padx': kwargs.pop('padx', 1)})
    else:
        _ = kwargs.pop('padx', 1)
    label = Label(iframe, **lkwargs)
    # print(label.configure())
    label.pack(side='left')
    # entry
    entry_text = tk.StringVar()
    if text is None:
        tt = 'None'
    elif isinstance(text, bool):
        if text:
            tt = 'True'
        else:
            tt = 'False'
    else:
        tt = str(text)
    entry_text.set(tt)
    entry = Entry(iframe, textvariable=entry_text, **kwargs)
    if command is not None:
        if isinstance(command, (list, tuple)):
            com0 = command[0]
            if len(command) > 1:
                com1 = command[1]
            else:
                com1 = command[0]
        else:
            com0 = command
            com1 = command
        entry.bind('<FocusOut>', com0)    # tab or click
        entry.bind('<Return>', com1)      # return
        entry.bind('<Key-Return>', com1)  # return
        entry.bind('<KP_Enter>', com1)    # return of numeric keypad
    entry.pack(side='left')
    # tooltip
    if tooltip:
        ttip = tk.StringVar()
        ttip.set(tooltip)
        etip = Tooltip(entry, ttip)
        return iframe, entry_label, entry_text, ttip
    else:
        return iframe, entry_label, entry_text


def add_imagemenu(frame, label="", values=[], images=[], command=None,
                  tooltip="", **kwargs):
    """
    Add a left-aligned menu with menubuttons having text and images
    with a Label before.

    Parameters
    ----------
    frame : tk widget
        Parent widget
    label : str, optional
        Text that appears in front of the menu (default: "")
    values : list of str, optional
        The choices that will appear in the drop-down menu (default: [])
    images : list of str, optional
        The images before the choices that will appear in the drop-down menu
        (default: [])
    command : function, optional
        Handler function to be called if new menu entry chosen (default: None).
    tooltip : str, optional
        Tooltip appearing after one second when hovering over
        the menu (default: "" = no tooltip)
    **kwargs : option=value pairs, optional
        All other options will be passed to the main ttk.Menubutton
    tk.StringVar
        variable for the text of the tooltip, if given.

    Returns
    -------
    tk.StringVar, ttk.Menubutton
        variable for the text before the menu, main tt.Menubutton widget

    Examples
    --------
    >>> self.rowcmap = Frame(self)
    >>> self.rowcmap.pack(side='top', fill='x')
    >>> self.cmaplbl, self.cmap = add_imagemenu(
    ...     self.rowcmap, label="cmap", values=self.cmaps,
    ...     images=self.imaps, command=self.selected_cmap)
    >>> self.cmap['text']  = 'RdYlBu'
    >>> self.cmap['image'] = self.imaps[self.cmaps.index('RdYlBu')]

    """
    from functools import partial
    iframe = Frame(frame)
    estr  = 'Same number of values and images needed for add_imagemenu.'
    estr += ' values (' + str(len(values)) + '): ' + str(values)
    estr += ', images (' + str(len(images)) + '): ' + str(images)
    assert len(values) == len(images), estr
    mb_label = tk.StringVar()
    mb_label.set(label)
    label = Label(iframe, textvariable=mb_label)
    label.pack(side='left')
    mb = ttk.Menubutton(iframe, image=images[0], text=values[0],
                        compound='left')
    sb = tk.Menu(mb, tearoff=False)
    mb.config(menu=sb)
    for i, v in enumerate(values):
        sb.add_command(label=v, image=images[i], compound='left',
                       command=partial(command, v))
    mb.pack(side='left')
    if tooltip:
        ttip = tk.StringVar()
        ttip.set(tooltip)
        mbtip = Tooltip(mb, ttip)
        return iframe, mb_label, mb, ttip
    else:
        return iframe, mb_label, mb


def add_menu(frame, label="", values=[], command=None, tooltip="", **kwargs):
    """
    Add a left-aligned menu with menubuttons with a Label before.

    Parameters
    ----------
    frame : CustomTkinter widget
        Parent widget
    label : str, optional
        Text that appears in front of the menu (default: "")
    values : list of str, optional
        The choices that will appear in the drop-down menu (default: [])
    command : function, optional
        Handler function to be called if new menu entry chosen (default: None).
    tooltip : str, optional
        Tooltip appearing after one second when hovering over
        the menu (default: "" = no tooltip)
    **kwargs : option=value pairs, optional
        All other options will be passed to the main Menubutton
    tk.StringVar
        variable for the text of the tooltip, if given.

    Returns
    -------
    tk.StringVar, Menubutton
        variable for the text before the menu, main tt.Menubutton widget

    Examples
    --------
    >>> self.rowzxy = Frame(self)
    >>> self.rowzxy.pack(side='top', fill='x')
    >>> self.xlbl, self.x = add_combobox(
    ...     self.rowzxy, label="x", values=columns, command=self.selected)

    """
    from functools import partial
    iframe = Frame(frame)
    mb_label = tk.StringVar()
    mb_label.set(label)
    label = Label(iframe, textvariable=mb_label)
    label.pack(side='left')
    if ihavectk:
        mb = Menubutton(iframe, values=values, command=command, **kwargs)
    else:
        mb = Menubutton(iframe, text=values[0], compound='left', **kwargs)
        sb = tk.Menu(mb, tearoff=False)
        mb.config(menu=sb)
        for i, v in enumerate(values):
            sb.add_command(label=v, compound='left',
                           command=partial(command, v))
    mb.pack(side='left')
    if tooltip:
        ttip = tk.StringVar()
        ttip.set(tooltip)
        mbtip = Tooltip(mb, ttip)
        return iframe, mb_label, mb, ttip
    else:
        return iframe, mb_label, mb


def add_scale(frame, label="", ini=0, tooltip="", **kwargs):
    """
    Add a left-aligned Scale with a Label before.

    Parameters
    ----------
    frame : CustomTkinter widget
        Parent widget
    label : str, optional
        Text that appears in front of the scale (default: "")
    ini : float, optional
        Initial value of scale (default: 0)
    tooltip : str, optional
        Tooltip appearing after one second when hovering over
        the scale (default: "" = no tooltip)
    **kwargs : option=value pairs, optional
        All other options will be passed to Scale

    Returns
    -------
    tk.StringVar, tk.DoubleVar, Scale
        variable for the text before the scale,
        value of scale,
        scale widget
    tk.StringVar
        variable for the text of the tooltip, if given.

    Examples
    --------
    >>> self.rowzxy = Frame(self)
    >>> self.rowzxy.pack(side='top', fill='x')
    >>> self.xlbl, self.x = add_scale(
    ...     self.rowzxy, label="x", values=columns, command=self.selected)

    """
    iframe = Frame(frame)
    s_label = tk.StringVar()
    s_label.set(label)
    label = Label(iframe, textvariable=s_label)
    label.pack(side='left')
    s_val = tk.DoubleVar()
    s_val.set(ini)
    if 'from_' not in kwargs:
        kwargs.update({'from_': 0})
    if 'to' not in kwargs:
        kwargs.update({'to': 100})
    if ihavectk:
        kwargs.update({'number_of_steps': kwargs['to'] - kwargs['from_'] + 1})
        length = kwargs.pop('length', -1)
        orient = kwargs.pop('orient', tk.HORIZONTAL)
        if length < 0:
            length = 100
        if orient == tk.HORIZONTAL:
            kwargs.update({'width': length})
        if orient == tk.VERTICAL:
            kwargs.update({'height': length})
    s = Scale(iframe, variable=s_val, **kwargs)
    s.pack(side='left')
    if tooltip:
        ttip = tk.StringVar()
        ttip.set(tooltip)
        stip = Tooltip(s, ttip)
        return iframe, s_label, s_val, s, ttip
    else:
        return iframe, s_label, s_val, s


def add_spinbox(frame, label="", values=[], command=None, tooltip="",
                **kwargs):
    """
    Add a left-aligned tk.Spinbox with a ttk.Label before.

    Parameters
    ----------
    frame : tk widget
        Parent widget
    label : str, optional
        Text that appears in front of the spinbox (default: "")
    values : list of str, optional
        The list of choices on the spinbox (default: [])
    command : function, optional
        Handler function bound to
        <Return>, '<Key-Return>', <KP_Enter>, and '<FocusOut>' (default: None).
    tooltip : str, optional
        Tooltip appearing after one second when hovering over
        the spinbox (default: "" = no tooltip)
    **kwargs : option=value pairs, optional
        All other options will be passed to tk.Spinbox

    Returns
    -------
    tk.StringVar, ttk.Label, tk.StringVar, tk.Spinbox
        variable for the text before the spinbox,
        label widget,
        variable for the text in the spinbox area,
        spinbox widget
    tk.StringVar
        variable for the text of the tooltip, if given.

    Examples
    --------
    >>> self.rowlev = ttk.Frame(self)
    >>> self.rowlev.pack(side='top', fill='x')
    >>> self.dlval, self.dl, self.dval, self.d = add_spinbox(
    ...     self.rowlev, label="dim", values=range(0,10),
    ...     command=self.spinned)

    """
    iframe = Frame(frame)
    width = kwargs.pop('width', 1)
    sbl_val = tk.StringVar()
    sbl_val.set(label)
    sbl = ttk.Label(iframe, textvariable=sbl_val)
    sbl.pack(side='left')
    sb_val = tk.StringVar()
    if len(values) > 0:
        sb_val.set(str(values[0]))
    sb = tk.Spinbox(iframe, values=values, command=command, width=width,
                    textvariable=sb_val, **kwargs)
    if command is not None:
        sb.bind('<Return>', command)      # return
        sb.bind('<Key-Return>', command)  # return
        sb.bind('<KP_Enter>', command)    # return of numeric keypad
        sb.bind('<FocusOut>', command)    # tab or click
    sb.pack(side='left')
    if tooltip:
        ttip = tk.StringVar()
        ttip.set(tooltip)
        sbtip = Tooltip(sb, ttip)
        return iframe, sbl_val, sbl, sb_val, sb, ttip
    else:
        return iframe, sbl_val, sbl, sb_val, sb


def add_tooltip(frame, tooltip="", **kwargs):
    """
    Add a writeable tooltip to a widget using the class Tooltip.

    Parameters
    ----------
    frame : tk widget
        Parent widget
    tooltip : str, optional
        Tooltip appearing after one second when hovering over
        the parent widget (default: "" = no tooltip)
    **kwargs : option=value pairs, optional
        All other options will be passed to tk.StringVar

    Returns
    -------
    tk.StringVar
        variable for the text of the tooltip.

    Examples
    --------
    >>> self.rowlev = ttk.Frame(self)
    >>> self.rowlev.pack(side='top', fill='x')
    >>> self.dlbl, self.dval, self.d = add_spinbox(
    ...     self.rowlev, label="dim", values=range(0,10),
    ...     command=self.spinned)
    >>> self.dtip = add_tooltip(self.d, 'Dimension')

    """
    ttip = tk.StringVar()
    ttip.set(tooltip)
    htip = Tooltip(frame, ttip)
    return ttip


# https://pythonassets.com/posts/scrollbar-in-tk-tkinter/
class Treeview(Frame):
    """
    Treeview class with optional horizontal and vertical scrollbars

    Examples
    --------
    Simple ttk.Treeview widget

    .. code-block:: python

       tree = Treeview(frame_widget)

    ttk.Treeview widget with vertical scrollbar

    .. code-block:: python

       tree = Treeview(frame_widget, yscroll=True)

    ttk.Treeview widget with vertical and horizontal scrollbars

    .. code-block:: python

       tree = Treeview(frame_widget, xscroll=True, yscroll=True)

    """
    def __init__(self, *args, xscroll=False, yscroll=False, **kwargs):
        super().__init__(*args, **kwargs)
        # scrollbars
        if xscroll:
            if ihavectk:
                self.hscrollbar = Scrollbar(self, orientation='horizontal')
            else:
                self.hscrollbar = Scrollbar(self, orient='horizontal')
        if yscroll:
            if ihavectk:
                self.vscrollbar = Scrollbar(self, orientation='vertical')
            else:
                self.vscrollbar = Scrollbar(self, orient='vertical')
        # treeview
        self.tv = ttk.Treeview(self)
        # pack scrollbars and treeview together
        if xscroll:
            self.tv.config(xscrollcommand=self.hscrollbar.set)
            self.hscrollbar.configure(command=self.tv.xview)
            self.hscrollbar.pack(side='bottom', fill='x')
        if yscroll:
            self.tv.config(yscrollcommand=self.vscrollbar.set)
            self.vscrollbar.configure(command=self.tv.yview)
            self.vscrollbar.pack(side='right', fill='y')
        self.tv.pack()
        # convenience functions
        self.tag_configure = self.tv.tag_configure
        self.config = self.tv.config
        self.column = self.tv.column
        self.heading = self.tv.heading
        self.insert = self.tv.insert



#!/usr/bin/env python
"""
Text widget for manipulation of DataFrame

This text widget allows putting in code to manipulate the current data frame.

This module was written by Matthias Cuntz while at Institut National de
Recherche pour l'Agriculture, l'Alimentation et l'Environnement (INRAE), Nancy,
France.

:copyright: Copyright 2023- Matthias Cuntz - mc (at) macu (dot) de
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following classes are provided:

.. autosummary::
   dfvTransform

History
   * Written Oct 2024 by Matthias Cuntz (mc (at) macu (dot) de)

"""
try:
    from customtkinter import CTkToplevel as Toplevel
    from customtkinter import CTkFrame as Frame
    from customtkinter import CTkButton as Button
    from customtkinter import CTkTextbox as Text
    from customtkinter import CTkScrollbar as Scrollbar
    ihavectk = True
except ModuleNotFoundError:
    from tkinter import Toplevel
    from tkinter.ttk import Frame
    from tkinter.ttk import Button
    from tkinter import Text
    from tkinter.ttk import Scrollbar
    ihavectk = False
from .ncvwidgets import add_tooltip


__all__ = ['dfvTransform']


class dfvTransform(Toplevel):
    """
    Window for reading Python code to change DataFrame.

    """

    #
    # Setup panel
    #

    def __init__(self, top, callback=None, **kwargs):
        super().__init__(top, **kwargs)

        self.top = top  # top window
        self.callback = callback

        self.name = 'dfvTransform'
        self.title("Manipulate DataFrame")
        if ihavectk:
            self.geometry('1000x540+100+10')
        else:
            self.geometry('1000x490+100+10')

        # copy for ease of use
        self.csvfile = self.top.csvfile
        self.newcsvfile = self.top.newcsvfile
        self.df = self.top.df
        self.sep = self.top.sep
        self.index_col = self.top.index_col
        self.skiprows = self.top.skiprows
        self.parse_dates = self.top.parse_dates
        self.date_format = self.top.date_format
        self.missing_value = self.top.missing_value
        self.cols = self.top.cols

        # 1. row - treeview current DataFrame
        self.rowtext = Frame(self)
        self.rowtext.pack(side='top', fill='x')
        if ihavectk:
            # px
            self.text = Text(self.rowtext, height=500, width=985,
                             font=("Helvetica", 18), wrap='none')
        else:
            # characters
            self.text = Text(self.rowtext, height=22, width=88,
                             font=("Helvetica", 18), wrap='none')
        # self.text = Text(self.rowtext, height=24, width=88,
        #                  font=("Helvetica", 18), wrap='none')
        self.vscroll = Scrollbar(self.rowtext, command=self.text.yview)
        self.hscroll = Scrollbar(self.rowtext, command=self.text.xview)
        self.text.configure(yscrollcommand=self.vscroll.set,
                            xscrollcommand=self.hscroll.set)
        self.text.insert('1.0', '# Example daily mean if datetime index\n')
        self.text.insert('2.0',
                         "self.df = self.df.resample('1D').mean().squeeze()")
        self.text.pack(side='left')
        self.vscroll.pack(side='right', fill='y')
        self.hscroll.pack(side='bottom', fill='x')

        # add cancel and read buttons to last row
        self.rowdone = Frame(self)
        self.rowdone.pack(side='top', fill='x')
        self.done = Button(self.rowdone, text="Transform",
                           command=self.exec_text)
        self.donetip = add_tooltip(self.done,
                                   'Execute DataFrame manipulations')
        self.done.pack(side='right', padx=10, pady=5)
        self.cancel = Button(self.rowdone, text="Cancel",
                             command=self.cancel)
        self.canceltip = add_tooltip(self.cancel,
                                     'Cancel DataFrame manipulation')
        self.cancel.pack(side='right', pady=5)

        self.focus_force()

    #
    # Event bindings
    #

    def cancel(self, event=None):
        if self.callback is not None:
            self.callback()
        # do not self.destroy() with ctk.CTkButton, leading to
        # 'invalid command name
        #     ".!dfvreadcsv.!ctkframe3.!ctkbutton2.!ctkcanvas"'
        # self.destroy() works with ttk.Button
        self.withdraw()

    def exec_text(self, event=None):
        tt = self.text.get('1.0', 'end')
        _ = exec(tt)
        self.top.df = self.df
        if self.callback is not None:
            self.callback()
        # do not self.destroy() with ctk.CTkButton, leading to
        # 'invalid command name
        #     ".!dfvreadcsv.!ctkframe3.!ctkbutton2.!ctkcanvas"'
        # self.destroy() works with ttk.Button
        self.withdraw()



#!/usr/bin/env python3
"""
Calling routine of dfvue

The calling routine sets up the toplevel root window and gets an
instance of the dfvMain class.

This module was written by Matthias Cuntz while at Institut National de
Recherche pour l'Agriculture, l'Alimentation et l'Environnement (INRAE), Nancy,
France.

:copyright: Copyright 2023- Matthias Cuntz - mc (at) macu (dot) de
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following functions are provided:

.. autosummary::
   dfvue

History
   * Written Jul 2023 by Matthias Cuntz (mc (at) macu (dot) de)
     adapted ncvue.py
   * Use CustomTkinter, Jun 2024, Matthias Cuntz
   * Use mix of grid and pack layout manager, Jun 2024, Matthias Cuntz
   * Use CustomTkinter only if installed, Jun 2024, Matthias Cuntz
   * Allow multiple input files, Oct 2024, Matthias Cuntz
   * Import pyplot for Windows, Oct 2024, Matthias Cuntz
   * Back to pack layout manager for resizing, Nov 2024, Matthias Cuntz
   * Pass Pandas DataFrame directly to dfvue, Jan 2025, Matthias Cuntz
   * Bugfix for checking if csvfile was given, Jan 2025, Matthias Cuntz
   * Use own ncvue-blue theme for customtkinter, Jan 2025, Matthias Cuntz

"""
import os
import platform
import sys
import tkinter as tk
import tkinter.ttk as ttk
try:
    import customtkinter
    from customtkinter import CTk as Tk
    from customtkinter import CTkToplevel as Toplevel
    ihavectk = True
except ModuleNotFoundError:
    from tkinter import Tk
    from tkinter import Toplevel
    ihavectk = False
from matplotlib import pyplot as plt
from .dfvmain import dfvMain


__all__ = ['dfvue']


def dfvue(df=None, csvfile='', sep='', index_col=None, skiprows=None,
          parse_dates=True, date_format=None, missing_value=None):
    """
    The main function to start the data frame GUI.

    Parameters
    ----------
    df : pandas.DataFrame, optional
        Pandas DataFrame will be used if no csvfile given
    csvfile : str or list of str, optional
        Name(s) of csv file (default: '').
    sep : str, optional
        Delimiter to use.
    index_col : str, optional
        Column(s) to use as index, either given as column index
        or string name.
    skiprows : str, optional
        Line numbers to skip (0-indexed, must include comma,
        e.g. "1," for skipping the second row) or
        number of lines to skip (int, without comma) at the start
        of the file.
    parse_dates : str, optional
        boolean. If True -> try parsing the index.

        list of int or names. e.g. If 1, 2, 3
        -> try parsing columns 1, 2, 3 each as a separate date column.

        list of lists. e.g. If [1, 3] -> combine columns 1 and 3 and
        parse as a single date column.

        dict, e.g. "foo" : [1, 3] -> parse columns 1, 3 as date and
        call result
    date_format : str, optional
        Will parse dates according to this format.
        For example: "%%Y-%%m-%%d %%H:%%M%%S". See
        https://docs.python.org/3/library/datetime.html#strftime-and-strptime-behavior
    missing_value : str, optional
        Missing or undefined value set to NaN.

    """
    # print(mpl.get_backend())
    ios = platform.system()  # Windows, Darwin, Linux
    if ios == 'Windows':
        # make Windows aware of high resolution displays
        # https://stackoverflow.com/questions/41315873/attempting-to-resolve-blurred-tkinter-text-scaling-on-windows-10-high-dpi-disp
        from ctypes import windll
        windll.shcore.SetProcessDpiAwareness(1)

    # Pyinstaller sets _MEIPASS if macOS app
    bundle_dir = getattr(sys, '_MEIPASS',
                         os.path.abspath(os.path.dirname(__file__)))

    top = Tk()
    top.withdraw()

    if not ihavectk:
        # style = ttk.Style()
        # print(style.theme_names(), style.theme_use())
        if ios == 'Darwin':
            theme = 'aqua'
            style = ttk.Style()
            try:
                style.theme_use(theme)
            except:
                pass
        elif ios == 'Windows':
            top.option_add("*Font", "Helvetica 10")
            plt.rc('font', size=13)
            # standard Windows themes
            # ('winnative', 'clam', 'alt', 'default', 'classic', 'vista',
            #  'xpnative')
            # theme = 'vista'
            # style = ttk.Style()
            # style.theme_use(theme)

            # 'azure' v2.x, 'sun-valley', 'forest' of rdbende
            top.tk.call('source', bundle_dir + '/themes/azure-2.0/azure.tcl')
            theme = 'light'  # light, dark
            top.tk.call("set_theme", theme)
        elif ios == 'Linux':
            # standard Linux schemes
            # theme = 'clam'  # 'clam', 'alt', 'default', 'classic'
            # style = ttk.Style()
            # style.theme_use(theme)

            # 'azure' v2.x, 'sun-valley', 'forest' of rdbende
            top.tk.call('source', bundle_dir + '/themes/azure-2.0/azure.tcl')
            theme = 'light'  # light, dark
            top.tk.call("set_theme", theme)

    if ihavectk:
        # customtkinter.set_default_color_theme("blue")
        # customtkinter.set_default_color_theme("dark-blue")
        # customtkinter.set_default_color_theme("green")
        # customtkinter.set_default_color_theme(
        #     f'{bundle_dir}/themes/customtkinter/dark-blue.json')
        customtkinter.set_default_color_theme(
            f'{bundle_dir}/themes/customtkinter/ncvue-blue.json')

    # set titlebar and taskbar icon only if "standalone",
    # i.e. not ipython or jupyter
    try:
        whichpy = get_ipython().__class__.__name__
    except NameError:
        whichpy = ''
    if not whichpy:
        icon = tk.PhotoImage(file=bundle_dir + '/images/dfvue_icon.png')
        top.iconphoto(True, icon)  # True: apply to all future toplevels
    else:
        icon = None

    root = Toplevel()
    root.name = 'dfvOne'
    if isinstance(csvfile, str):
        csvfile = [csvfile]
    if csvfile[0]:
        tit = f"dfvue {csvfile}"
    else:
        tit = "dfvue"
    root.title(tit)
    root.geometry('1000x800+110+0')

    # Connect csv file and add information to top
    top.os = ios           # operating system
    top.icon = icon        # app icon
    top.csvfile = csvfile  # file name or file handle
    top.newcsvfile = True  # new file after command line
    if csvfile[0]:
        top.newcsvfile = False
    top.df = df            # pandas.DataFrame of csvfile
    # variable list
    if df is not None:
        top.newcsvfile = False
        rows = top.df.shape[0]
        top.cols = [ f'{cc} ({rows} {top.df[cc].dtype.name})'
                     for cc in top.df.columns ]
    else:
        top.cols = []
    top.sep = sep
    top.index_col = index_col
    top.skiprows = skiprows
    top.parse_dates = parse_dates
    top.date_format = date_format
    top.missing_value = missing_value
    root.top = top

    def on_closing():
        top.quit()
        top.destroy()
    root.protocol("WM_DELETE_WINDOW", on_closing)

    # 1st plotting window
    main_frame = dfvMain(root)
    main_frame.pack(fill=tk.BOTH, expand=1)

    top.mainloop()



#!/usr/bin/env python
"""
Scatter/Line panel of dfvue.

The panel allows plotting variables against time or two variables against
each other. A second variable can be plotted in the same graph using the
right-hand-side y-axis.

This module was written by Matthias Cuntz while at Institut National de
Recherche pour l'Agriculture, l'Alimentation et l'Environnement (INRAE), Nancy,
France.

:copyright: Copyright 2023- Matthias Cuntz - mc (at) macu (dot) de
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following classes are provided:

.. autosummary::
   dfvScatter

History
   * Written Jul 2023 by Matthias Cuntz (mc (at) macu (dot) de)
   * Use CustomTkinter, Jun 2024, Matthias Cuntz
   * Use mix of grid and pack layout manager, Jun 2024, Matthias Cuntz
   * Use CustomTkinter only if installed, Jun 2024, Matthias Cuntz
   * Allow multiple input files, Oct 2024, Matthias Cuntz
   * Back to pack layout manager for resizing, Nov 2024, Matthias Cuntz
   * Bugfix for checking if csvfile was given, Jan 2025, Matthias Cuntz
   * Removed addition of index to column names in sortvars,
     Jan 2025, Matthias Cuntz
   * Add xlim, ylim, and y2lim options, Jan 2025, Matthias Cuntz

"""
import tkinter as tk
try:
    from customtkinter import CTkFrame as Frame
    from customtkinter import CTkButton as Button
    from customtkinter import CTkLabel as Label
    from customtkinter import CTkComboBox as Combobox
    ihavectk = True
except ModuleNotFoundError:
    from tkinter.ttk import Frame
    from tkinter.ttk import Button
    from tkinter.ttk import Label
    from tkinter.ttk import Combobox
    ihavectk = False
from tkinter import filedialog
import warnings
import numpy as np
import pandas as pd
from .dfvutils import clone_dfvmain, format_coord_scatter, vardim2var
from .dfvutils import parse_entry
from .ncvwidgets import add_checkbutton, add_combobox, add_entry
from .ncvwidgets import add_tooltip
from .dfvreadcsv import dfvReadcsv
from .dfvtransform import dfvTransform
from matplotlib import pyplot as plt
try:
    plt.style.use('seaborn-v0_8-dark')
except OSError:
    plt.style.use('seaborn-dark')


__all__ = ['dfvScatter']


def _minmax_ylim(ylim, ylim2):
    """
    Get minimum of first elements of lists `ylim` and `ylim2` and
    maximum of second element of the two lists.

    Returns minimum, maximum.

    """
    if (ylim[0] is not None) and (ylim2[0] is not None):
        ymin = min(ylim[0], ylim2[0])
    else:
        if (ylim[0] is not None):
            ymin = ylim[0]
        else:
            ymin = ylim2[0]
    if (ylim[1] is not None) and (ylim2[1] is not None):
        ymax = max(ylim[1], ylim2[1])
    else:
        if (ylim[1] is not None):
            ymax = ylim[1]
        else:
            ymax = ylim2[1]
    return ymin, ymax


class dfvScatter(Frame):
    """
    Panel for scatter and line plots.

    Sets up the layout with the figure canvas, variable selectors, dimension
    spinboxes, and options.

    Contains various commands that manage what will be drawn or redrawn if
    something is selected, changed, checked, etc.

    Contains three drawing routines. `redraw_y` and `redraw_y2` redraw the
    y-axes without changing zoom level, etc. `redraw` is called if a new
    x-variable was selected or the `Redraw`-button was pressed. It resets
    all axes, resetting zoom, etc.

    """

    #
    # Setup panel
    #

    def __init__(self, master, **kwargs):
        from functools import partial
        from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
        from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk
        from matplotlib.figure import Figure

        super().__init__(master, **kwargs)

        self.name   = 'Scatter/Line'
        self.master = master
        self.top    = master.top

        # copy for ease of use
        self.csvfile = self.top.csvfile
        self.newcsvfile = self.top.newcsvfile
        self.df = self.top.df
        self.sep = self.top.sep
        self.index_col = self.top.index_col
        self.skiprows = self.top.skiprows
        self.parse_dates = self.top.parse_dates
        self.date_format = self.top.date_format
        self.missing_value = self.top.missing_value
        self.cols = self.top.cols

        # selections and options
        columns = [''] + self.cols
        # colors
        c = list(plt.rcParams['axes.prop_cycle'])
        col1 = c[0]['color']  # blue
        col2 = c[3]['color']  # red
        # color tooltip
        ctstr = ("- color names: red, green, blue, yellow, ...\n"
                 "- single characters: b (blue), g (green), r (red), c (cyan),"
                 " m (magenta), y (yellow), k (black), w (white)\n"
                 "- hex RGB: #rrggbb such such as #ff9300 (orange)\n"
                 "- gray level: float between 0 and 1\n"
                 "- RGA (red, green, blue) or RGBA (red, green, blue, alpha)"
                 " tuples between 0 and 1, e.g. (1, 0.57, 0) for orange\n"
                 "- name from xkcd color survey, e.g. xkcd:sky blue")
        # marker tooltip
        mtstr = (". (point), ',' (pixel), o (circle),\n"
                 "v (triangle_down), ^ (triangle_up),\n"
                 "< (triangle_left), > (triangle_right),\n"
                 "1 (tri_down), 2 (tri_up), 3 (tri_left), 4 (tri_right),"
                 " 8 (octagon),\n"
                 "s (square), p (pentagon), P (plus (filled)),\n"
                 "* (star), h (hexagon1), H (hexagon2),\n"
                 "+ (plus), x (x), X (x (filled)),\n"
                 "D (diamond), d (thin_diamond),\n"
                 "| (vline), _ (hline), or None")
        if ihavectk:
            # width of combo boxes in px
            combowidth = 288
            # widths of entry widgets in px
            ewsmall = 20
            ewmed = 45
            ewbig = 70
            ew2big = 100
        else:
            # width of combo boxes in characters
            combowidth = 35  # 25
            # widths of entry widgets in characters
            ewsmall = 3
            ewmed = 5  # 4
            ewbig = 8  # 7
            ew2big = 10

        # open file and new window
        self.rowwin = Frame(self)
        self.rowwin.pack(side=tk.TOP, fill=tk.X)
        self.newfile = Button(self.rowwin, text="Open File",
                              command=self.new_csv)
        self.newfiletip = add_tooltip(self.newfile, 'Open a new csv file')
        self.newfile.pack(side=tk.LEFT)
        self.newwin = Button(
            self.rowwin, text="New Window",
            command=partial(clone_dfvmain, self.master))
        self.newwintip = add_tooltip(
            self.newwin, 'Open secondary dfvue window')
        self.newwin.pack(side=tk.RIGHT)

        # plotting canvas
        self.figure = Figure(facecolor="white", figsize=(1, 1))
        self.axes   = self.figure.add_subplot(111)
        self.axes2  = self.axes.twinx()
        self.axes2.yaxis.set_label_position("right")
        self.axes2.yaxis.tick_right()
        self.canvas = FigureCanvasTkAgg(self.figure, master=self)
        self.canvas.draw()
        self.tkcanvas = self.canvas.get_tk_widget()
        self.tkcanvas.pack(side=tk.TOP, fill=tk.BOTH, expand=1)

        # matplotlib toolbar
        # toolbar uses pack internally -> put into frame
        self.toolbar = NavigationToolbar2Tk(self.canvas, self,
                                            pack_toolbar=True)
        self.toolbar.update()
        self.toolbar.pack(side=tk.TOP, fill=tk.X)

        # x-axis and left y-axis
        self.rowxy = Frame(self)
        self.rowxy.pack(side=tk.TOP, fill=tk.X)
        # block with x
        self.blockx = Frame(self.rowxy)
        self.blockx.pack(side=tk.LEFT)
        # x
        self.xframe, self.xlbl, self.x, self.xtip = add_combobox(
            self.blockx, label="x", values=columns, command=self.selected_x,
            width=combowidth,
            tooltip="Choose variable of x-axis.\nTake index if 'None' (fast).")
        self.xframe.pack(side=tk.LEFT)
        # invert x
        (self.inv_xframe, self.inv_xlbl, self.inv_x,
         self.inv_xtip) = add_checkbutton(
             self.blockx, label="invert x", value=False,
             command=self.checked_x,
             tooltip="Invert x-axis")
        self.inv_xframe.pack(side=tk.LEFT)
        # space1
        self.space1lbl = tk.StringVar()
        self.space1lbl.set("   ")
        space1lab = Label(self.rowxy, textvariable=self.space1lbl)
        space1lab.pack(side='left')
        # block with y
        self.blocky = Frame(self.rowxy)
        self.blocky.pack(side=tk.LEFT)
        # y label
        self.ylbl = tk.StringVar()
        self.ylbl.set("y")
        ylab = Label(self.blocky, textvariable=self.ylbl)
        ylab.pack(side='left')
        # previous and next buttons
        self.bprev_y = Button(self.blocky, text="<", width=1,
                              command=self.prev_y)
        self.bprev_ytip = add_tooltip(self.bprev_y, 'Previous variable')
        self.bprev_y.pack(side='left')
        self.bnext_y = Button(self.blocky, text=">", width=1,
                              command=self.next_y)
        self.bnext_ytip = add_tooltip(self.bnext_y, 'Next variable')
        self.bnext_y.pack(side='left')
        # y
        if ihavectk:
            self.y = Combobox(self.blocky, values=columns, width=combowidth,
                              command=self.selected_y)
        else:
            self.y = Combobox(self.blocky, values=columns, width=combowidth)
            self.y.bind("<<ComboboxSelected>>", self.selected_y)
        self.ytip = add_tooltip(self.y, 'Choose variable of y-axis')
        self.y.pack(side='left')
        # invert y
        self.line_y = []
        (self.inv_yframe, self.inv_ylbl, self.inv_y,
         self.inv_ytip) = add_checkbutton(
             self.blocky, label="invert y", value=False,
             command=self.checked_y,
             tooltip="Invert y-axis")
        self.inv_yframe.pack(side=tk.LEFT)
        # redraw button
        self.bredraw = Button(self.rowxy, text="Redraw",
                              command=self.redraw)
        self.bredrawtip = add_tooltip(self.bredraw, 'Redraw, resetting zoom')
        self.bredraw.pack(side=tk.RIGHT)

        # options for lhs y-axis
        self.rowxyopt = Frame(self)
        self.rowxyopt.pack(side=tk.TOP, fill=tk.X)
        self.blockyopt = Frame(self.rowxyopt)
        self.blockyopt.pack(side=tk.LEFT)
        self.lsframe, self.lslbl, self.ls, self.lstip = add_entry(
            self.blockyopt, label="ls", text='-', width=ewmed,
            command=self.entered_y,
            tooltip="Line style: -, --, -., :, or None")
        self.lsframe.pack(side=tk.LEFT)
        self.lwframe, self.lwlbl, self.lw, self.lwtip = add_entry(
            self.blockyopt, label="lw", text='1', width=ewsmall,
            command=self.entered_y, tooltip="Line width")
        self.lwframe.pack(side=tk.LEFT)
        self.lcframe, self.lclbl, self.lc, self.lctip = add_entry(
            self.blockyopt, label="c", text=col1, width=ewbig,
            command=self.entered_y,
            tooltip="Line color:\n" + ctstr)
        self.lcframe.pack(side=tk.LEFT)
        (self.markerframe, self.markerlbl, self.marker,
         self.markertip) = add_entry(
            self.blockyopt, label="marker", text='None', width=ewmed,
            command=self.entered_y,
            tooltip="Marker symbol:\n" + mtstr)
        self.markerframe.pack(side=tk.LEFT)
        self.msframe, self.mslbl, self.ms, self.mstip = add_entry(
            self.blockyopt, label="ms", text='1', width=ewsmall,
            command=self.entered_y, tooltip="Marker size")
        self.msframe.pack(side=tk.LEFT)
        self.mfcframe, self.mfclbl, self.mfc, self.mfctip = add_entry(
            self.blockyopt, label="mfc", text=col1, width=ewbig,
            command=self.entered_y,
            tooltip="Marker fill color:\n" + ctstr)
        self.mfcframe.pack(side=tk.LEFT)
        self.mecframe, self.meclbl, self.mec, self.mectip = add_entry(
            self.blockyopt, label="mec", text=col1, width=ewbig,
            command=self.entered_y,
            tooltip="Marker edge color:\n" + ctstr)
        self.mecframe.pack(side=tk.LEFT)
        self.mewframe, self.mewlbl, self.mew, self.mewtip = add_entry(
            self.blockyopt, label="mew", text='1', width=ewsmall,
            command=self.entered_y, tooltip="Marker edge width")
        self.mewframe.pack(side=tk.LEFT)
        self.space2lbl = tk.StringVar()
        self.space2lbl.set(" ")
        space2lab = Label(self.blockyopt, textvariable=self.space2lbl)
        space2lab.pack(side='left')
        self.xlimframe, self.xlimlbl, self.xlim, self.xlimtip = add_entry(
            self.blockyopt, label="xlim", text='None', width=ew2big,
            command=self.entered_y,
            tooltip="xmin, xmax\nSet to None for free scaling.")
        self.xlimframe.pack(side=tk.LEFT)
        self.space3lbl = tk.StringVar()
        self.space3lbl.set(" ")
        space3lab = Label(self.blockyopt, textvariable=self.space3lbl)
        space3lab.pack(side='left')
        self.ylimframe, self.ylimlbl, self.ylim, self.ylimtip = add_entry(
            self.blockyopt, label="ylim", text='None', width=ew2big,
            command=self.entered_y,
            tooltip="ymin, ymax\nSet to None for free scaling.")
        self.ylimframe.pack(side=tk.LEFT)

        # redraw button
        self.bsort = Button(self.rowxyopt, text="Sort vars",
                            command=self.sortvars)
        self.bsorttip = add_tooltip(self.bsort, 'Sort variable names')
        self.bsort.pack(side=tk.RIGHT)

        # Transform data frame
        self.rowtransform = Frame(self)
        self.rowtransform.pack(side=tk.TOP, fill=tk.X)
        self.transform = Button(self.rowtransform, text="Transform df",
                                command=self.transform_df)
        self.transformtip = add_tooltip(self.transform, 'Manipulate DataFrame')
        self.transform.pack(side=tk.RIGHT)

        # right y2-axis
        self.rowyy2 = Frame(self)
        self.rowyy2.pack(side=tk.TOP, fill=tk.X)
        self.blocky2 = Frame(self.rowyy2)
        self.blocky2.pack(side=tk.TOP, fill=tk.X)
        # y label
        self.ylbl2 = tk.StringVar()
        self.ylbl2.set("y2")
        ylab2 = Label(self.blocky2, textvariable=self.ylbl2)
        ylab2.pack(side='left')
        # previous and next buttons
        self.bprev_y2 = Button(self.blocky2, text="<", width=1,
                               command=self.prev_y2)
        self.bprev_y2tip = add_tooltip(self.bprev_y2, 'Previous variable')
        self.bprev_y2.pack(side='left')
        self.bnext_y2 = Button(self.blocky2, text=">", width=1,
                               command=self.next_y2)
        self.bnext_y2tip = add_tooltip(self.bnext_y2, 'Next variable')
        self.bnext_y2.pack(side='left')
        # y
        if ihavectk:
            self.y2 = Combobox(self.blocky2, values=columns, width=combowidth,
                               command=self.selected_y2)
        else:
            self.y2 = Combobox(self.blocky2, values=columns, width=combowidth)
            self.y2.bind("<<ComboboxSelected>>", self.selected_y2)
        self.y2tip = add_tooltip(self.y2,
                                 'Choose variable of right-hand-side y-axis')
        self.y2.pack(side='left')
        # invert y2
        self.line_y2 = []
        (self.inv_y2frame, self.inv_y2lbl, self.inv_y2,
         self.inv_y2tip) = add_checkbutton(
             self.blocky2, label="invert y2", value=False,
             command=self.checked_y2,
             tooltip="Invert right-hand-side y-axis")
        self.inv_y2frame.pack(side=tk.LEFT)
        tstr = "Same limits for left-hand-side and right-hand-side y-axes"
        (self.same_yframe, self.same_ylbl, self.same_y,
         self.same_ytip) = add_checkbutton(
             self.blocky2, label="same y-axes", value=False,
             command=self.checked_yy2, tooltip=tstr)
        self.same_yframe.pack(side=tk.LEFT)

        # options for rhs y-axis 2
        self.rowy2opt = Frame(self)
        self.rowy2opt.pack(side=tk.TOP, fill=tk.X)
        self.blocky2opt = Frame(self.rowy2opt)
        self.blocky2opt.pack(side=tk.TOP, fill=tk.X)
        self.ls2frame, self.ls2lbl, self.ls2, self.ls2tip = add_entry(
            self.blocky2opt, label="ls", text='-', width=ewmed,
            command=self.entered_y2,
            tooltip="Line style: -, --, -., :, or None")
        self.ls2frame.pack(side=tk.LEFT)
        self.lw2frame, self.lw2lbl, self.lw2, self.lw2tip = add_entry(
            self.blocky2opt, label="lw", text='1', width=ewsmall,
            command=self.entered_y2, tooltip="Line width")
        self.lw2frame.pack(side=tk.LEFT)
        self.lc2frame, self.lc2lbl, self.lc2, self.lc2tip = add_entry(
            self.blocky2opt, label="c", text=col2, width=ewbig,
            command=self.entered_y2,
            tooltip="Line color:\n" + ctstr)
        self.lc2frame.pack(side=tk.LEFT)
        (self.marker2frame, self.marker2lbl, self.marker2,
         self.marker2tip) = add_entry(
            self.blocky2opt, label="marker", text='None', width=ewmed,
            command=self.entered_y2,
            tooltip="Marker symbol:\n" + mtstr)
        self.marker2frame.pack(side=tk.LEFT)
        self.ms2frame, self.ms2lbl, self.ms2, self.ms2tip = add_entry(
            self.blocky2opt, label="ms", text='1', width=ewsmall,
            command=self.entered_y2, tooltip="Marker size")
        self.ms2frame.pack(side=tk.LEFT)
        self.mfc2frame, self.mfc2lbl, self.mfc2, self.mfc2tip = add_entry(
            self.blocky2opt, label="mfc", text=col2, width=ewbig,
            command=self.entered_y2, tooltip="Marker fill color:\n" + ctstr)
        self.mfc2frame.pack(side=tk.LEFT)
        self.mec2frame, self.mec2lbl, self.mec2, self.mec2tip = add_entry(
            self.blocky2opt, label="mec", text=col2, width=ewbig,
            command=self.entered_y2, tooltip="Marker edge color:\n" + ctstr)
        self.mec2frame.pack(side=tk.LEFT)
        self.mew2frame, self.mew2lbl, self.mew2, self.mew2tip = add_entry(
            self.blocky2opt, label="mew", text='1', width=ewsmall,
            command=self.entered_y2, tooltip="Marker edge width")
        self.mew2frame.pack(side=tk.LEFT)
        self.space4lbl = tk.StringVar()
        self.space4lbl.set(" ")
        space4lab = Label(self.blocky2opt, textvariable=self.space4lbl)
        space4lab.pack(side='left')
        self.y2limframe, self.y2limlbl, self.y2lim, self.y2limtip = add_entry(
            self.blocky2opt, label="y2lim", text='None', width=ew2big,
            command=self.entered_y2,
            tooltip="y2min, y2max\nSet to None for free scaling.")
        self.y2limframe.pack(side=tk.LEFT)

        # Quit button
        self.rowquit = Frame(self)
        self.rowquit.pack(side=tk.TOP, fill=tk.X)
        self.bquit = Button(self.rowquit, text="Quit",
                            command=self.master.top.destroy)
        self.bquittip = add_tooltip(self.bquit, 'Quit dfvue')
        self.bquit.pack(side=tk.RIGHT, fill=tk.X)

        if self.csvfile[0] and (self.master.master.name == 'dfvOne'):
            self.new_df()

    #
    # Event bindings
    #

    def checked_x(self):
        """
        Command called if any checkbutton for x-axis was checked or unchecked.

        Redraws left-hand-side and right-hand-side y-axes.

        """
        self.redraw_y()
        self.redraw_y2()

    def checked_y(self):
        """
        Command called if any checkbutton for left-hand-side y-axis was checked
        or unchecked.

        Redraws left-hand-side y-axis.

        """
        self.redraw_y()

    def checked_y2(self):
        """
        Command called if any checkbutton for right-hand-side y-axis was
        checked or unchecked.

        Redraws right-hand-side y-axis.

        """
        self.redraw_y2()

    def checked_yy2(self):
        """
        Command called if any checkbutton was checked or unchecked that
        concerns both, the left-hand-side and right-hand-side y-axes.

        Redraws left-hand-side and right-hand-side y-axes.

        """
        self.ylim.set('None')
        self.y2lim.set('None')
        self.redraw_y()
        self.redraw_y2()

    def entered_y(self, event):
        """
        Command called if option was entered for left-hand-side y-axis.

        Redraws left-hand-side y-axis.

        """
        self.redraw_y()

    def entered_y2(self, event):
        """
        Command called if option was entered for right-hand-side y-axis.

        Redraws right-hand-side y-axis.

        """
        self.redraw_y2()

    def new_csv(self):
        """
        Open a new csv file and connect it to top.

        """
        # get new csv file name
        self.top.csvfile = filedialog.askopenfilename(
            parent=self, title='Choose csv file(s)', multiple=True)

        if self.top.csvfile[0]:
            self.top.newcsvfile = True
            self.new_df()

    def new_df(self):
        """
        Read new DataFrame.

        """
        with warnings.catch_warnings():
            warnings.simplefilter(action='ignore', category=FutureWarning)
            self.top.df = pd.read_csv(self.top.csvfile[0], nrows=40)
        self.readcsvwin = dfvReadcsv(self.top, callback=self.reset)

    def next_y(self):
        """
        Command called if next button for the left-hand-side y-variable was
        pressed.

        Resets dimensions of left-hand-side y-variable.
        Redraws plot.

        """
        y = self.y.get()
        if ihavectk:
            cols = self.y.cget("values")
        else:
            cols = self.y["values"]
        idx = cols.index(y)
        idx += 1
        if idx < len(cols):
            self.y.set(cols[idx])
            self.ylim.set('None')
            self.redraw()

    def next_y2(self):
        """
        Command called if next button for the right-hand-side y-variable was
        pressed.

        Resets dimensions of right-hand-side y-variable.
        Redraws plot.

        """
        y2 = self.y2.get()
        if ihavectk:
            cols = self.y2.cget("values")
        else:
            cols = self.y2["values"]
        idx = cols.index(y2)
        idx += 1
        if idx < len(cols):
            self.y2.set(cols[idx])
            self.y2lim.set('None')
            self.redraw()

    # def onpick(self, event):
    #     print('in pick')
    #     print('you pressed', event.button, event.xdata, event.ydata)
    #     thisline = event.artist
    #     xdata = thisline.get_xdata()
    #     ydata = thisline.get_ydata()
    #     ind = event.ind
    #     points = tuple(zip(xdata[ind], ydata[ind]))
    #     print('onpick points:', points)

    def prev_y(self):
        """
        Command called if previous button for the left-hand-side y-variable was
        pressed.

        Resets dimensions of left-hand-side y-variable.
        Redraws plot.

        """
        y = self.y.get()
        if ihavectk:
            cols = self.y.cget("values")
        else:
            cols = self.y["values"]
        idx = cols.index(y)
        idx -= 1
        if idx > 0:
            self.y.set(cols[idx])
            self.ylim.set('None')
            self.redraw()

    def prev_y2(self):
        """
        Command called if previous button for the right-hand-side
        y-variable was pressed.

        Resets dimensions of right-hand-side y-variable.
        Redraws plot.

        """
        y2 = self.y2.get()
        if ihavectk:
            cols = self.y2.cget("values")
        else:
            cols = self.y2["values"]
        idx = cols.index(y2)
        idx -= 1
        if idx > 0:
            self.y2.set(cols[idx])
            self.y2lim.set('None')
            self.redraw()

    def selected_x(self, event):
        """
        Command called if x-variable was selected with combobox.

        Triggering `event` was bound to the combobox.

        Redraws plot.

        """
        self.xlim.set('None')
        self.redraw()

    def selected_y(self, event):
        """
        Command called if left-hand-side y-variable was selected with
        combobox.

        Triggering `event` was bound to the combobox.

        Redraws plot.

        """
        self.ylim.set('None')
        self.redraw()

    def selected_y2(self, event):
        """
        Command called if right-hand-side y-variable was selected with
        combobox.

        Triggering `event` was bound to the combobox.

        Redraws plot.

        """
        self.y2lim.set('None')
        self.redraw()

    def sortvars(self):
        """
        Sort variable names in comboboxes.

        Triggering `event` was bound to button bsort.

        """
        if self.df is not None:
            columns = list(self.df.columns)
            columns.sort()
            rows = self.df.shape[0]
            columns = [ f'{cc} ({rows} {self.df[cc].dtype.name})'
                        for cc in columns ]
            # if self.df.index.name is not None:
            #     idx = (f'{self.df.index.name}'
            #            f' (index {rows} {self.df.index.dtype.name})')
            # else:
            #     idx = f'index ({rows} {self.df.index.dtype.name})'
            # self.cols = [idx]
            # self.cols.extend(columns)
            self.cols = columns
            # set variables
            columns = [''] + self.cols
            x = self.x.get()
            if ihavectk:
                self.x.configure(values=columns)
            else:
                self.x['values'] = columns
            self.x.set(x)
            y = self.y.get()
            if ihavectk:
                self.y.configure(values=columns)
            else:
                self.y['values'] = columns
            self.y.set(y)
            y2 = self.y2.get()
            if ihavectk:
                self.y2.configure(values=columns)
            else:
                self.y2['values'] = columns
            self.y2.set(y2)

    def transform_df(self):
        """
        Manipulate DataFrame.

        """
        self.df = dfvTransform(self.top, callback=self.reset)

    #
    # Methods
    #

    def reinit(self):
        """
        Reinitialise the panel from top.

        """
        # reinit from top
        self.csvfile = self.top.csvfile
        self.newcsvfile = self.top.newcsvfile
        self.df = self.top.df
        self.sep = self.top.sep
        self.index_col = self.top.index_col
        self.skiprows = self.top.skiprows
        self.parse_dates = self.top.parse_dates
        self.date_format = self.top.date_format
        self.missing_value = self.top.missing_value
        self.cols = self.top.cols
        if self.top.csvfile[0]:
            tit = f"dfvue {self.top.csvfile}"
        else:
            tit = "dfvue"
        self.master.master.title(tit)
        # set variables
        columns = [''] + self.cols
        if ihavectk:
            self.x.configure(values=columns)
        else:
            self.x['values'] = columns
        self.x.set(columns[0])
        self.xlim.set('None')
        if ihavectk:
            self.y.configure(values=columns)
        else:
            self.y['values'] = columns
        self.y.set(columns[0])
        self.ylim.set('None')
        if ihavectk:
            self.y2.configure(values=columns)
        else:
            self.y2['values'] = columns
        self.y2.set(columns[0])
        self.y2lim.set('None')

    def reset(self):
        """
        Reinit and redraw.

        """
        self.reinit()
        self.redraw()

    #
    # Plot
    #

    def redraw_y(self):
        """
        Redraw the left-hand-side y-axis.

        Reads left-hand-side `y` variable name, the current settings of
        its dimension spinboxes, as well as all other plotting options.
        Then redraws the left-hand-side y-axis.

        """
        # get all states
        # rowxy
        y = self.y.get()
        if y != '':
            inv_y = self.inv_y.get()
            ylim = parse_entry(self.ylim.get())
            ylim2 = parse_entry(self.y2lim.get())
            # rowxyopt
            ls = self.ls.get()
            lw = float(self.lw.get())
            c = str(self.lc.get())
            try:
                if isinstance(eval(c), tuple):
                    c = eval(c)
            except:  # several different exceptions possible
                pass
            m = self.marker.get()
            ms = float(self.ms.get())
            mfc = self.mfc.get()
            try:
                if isinstance(eval(mfc), tuple):
                    mfc = eval(mfc)
            except:
                pass
            mec = self.mec.get()
            try:
                if isinstance(eval(mec), tuple):
                    mec = eval(mec)
            except:
                pass
            mew = float(self.mew.get())
            # rowy2
            y2 = self.y2.get()
            same_y = self.same_y.get()
            # y plotting styles
            pargs = {'linestyle': ls,
                     'linewidth': lw,
                     'marker': m,
                     'markersize': ms,
                     'markerfacecolor': mfc,
                     'markeredgecolor': mec,
                     'markeredgewidth': mew}
            vy = vardim2var(y)
            ylab = self.df[vy].name
            if len(self.line_y) == 1:
                # set color only if single line,
                # None and 'None' do not work for multiple lines
                pargs['color'] = c
            # set style
            for ll in self.line_y:
                plt.setp(ll, **pargs)
            if 'color' in pargs:
                ic = pargs['color']
                if (ic != 'None'):
                    self.axes.spines['left'].set_color(ic)
                    self.axes.tick_params(axis='y', colors=ic)
                    self.axes.yaxis.label.set_color(ic)
            self.axes.yaxis.set_label_text(ylab)
            # same y-axes
            if not isinstance(ylim, list):
                ylim = self.axes.get_ylim()
            if not isinstance(ylim2, list):
                ylim2 = self.axes2.get_ylim()
            if same_y and (y2 != ''):
                ymin, ymax = _minmax_ylim(ylim, ylim2)
                if (ymin is not None) and (ymax is not None):
                    ylim  = [ymin, ymax]
                    ylim2 = [ymin, ymax]
                    self.axes.set_ylim(ylim)
                    self.axes2.set_ylim(ylim2)
            # invert y-axis
            if inv_y and (ylim[0] is not None):
                if ylim[0] < ylim[1]:
                    ylim = ylim[::-1]
                self.axes.set_ylim(ylim)
            else:
                if ylim[1] < ylim[0]:
                    ylim = ylim[::-1]
                self.axes.set_ylim(ylim)
            # invert x-axis
            inv_x = self.inv_x.get()
            xlim = parse_entry(self.xlim.get())
            if not isinstance(xlim, list):
                xlim = self.axes.get_xlim()
            if inv_x and (xlim[0] is not None):
                if xlim[0] < xlim[1]:
                    xlim = xlim[::-1]
                self.axes.set_xlim(xlim)
            else:
                if xlim[1] < xlim[0]:
                    xlim = xlim[::-1]
                self.axes.set_xlim(xlim)
            # redraw
            self.canvas.draw()
            self.toolbar.update()

    def redraw_y2(self):
        """
        Redraw the right-hand-side y-axis.

        Reads right-hand-side `y` variable name, the current settings of
        its dimension spinboxes, as well as all other plotting options.
        Then redraws the right-hand-side y-axis.

        """
        # get all states
        # rowy2
        y2 = self.y2.get()
        if y2 != '':
            # # rowxy
            # y = self.y.get()
            # rowy2
            inv_y2 = self.inv_y2.get()
            same_y = self.same_y.get()
            ylim = parse_entry(self.ylim.get())
            ylim2 = parse_entry(self.y2lim.get())
            # rowy2opt
            ls = self.ls2.get()
            lw = float(self.lw2.get())
            c = self.lc2.get()
            try:
                if isinstance(eval(c), tuple):
                    c = eval(c)
            except:
                pass
            m = self.marker2.get()
            ms = float(self.ms2.get())
            mfc = self.mfc2.get()
            try:
                if isinstance(eval(mfc), tuple):
                    mfc = eval(mfc)
            except:
                pass
            mec = self.mec2.get()
            try:
                if isinstance(eval(mec), tuple):
                    mec = eval(mec)
            except:
                pass
            mew = float(self.mew2.get())
            # y plotting styles
            pargs = {'linestyle': ls,
                     'linewidth': lw,
                     'marker': m,
                     'markersize': ms,
                     'markerfacecolor': mfc,
                     'markeredgecolor': mec,
                     'markeredgewidth': mew}
            vy2 = vardim2var(y2)
            ylab = self.df[vy2].name
            if len(self.line_y2) == 1:
                # set color only if single line,
                # None and 'None' do not work for multiple lines
                pargs['color'] = c
            # set style
            for ll in self.line_y2:
                plt.setp(ll, **pargs)
            if 'color' in pargs:
                ic = pargs['color']
                if (ic != 'None'):
                    self.axes2.spines['left'].set_color(ic)
                    self.axes2.tick_params(axis='y', colors=ic)
                    self.axes2.yaxis.label.set_color(ic)
            self.axes2.yaxis.set_label_text(ylab)
            # same y-axes
            if not isinstance(ylim, list):
                ylim = self.axes.get_ylim()
            if not isinstance(ylim2, list):
                ylim2 = self.axes2.get_ylim()
            if same_y and (y2 != ''):
                ymin, ymax = _minmax_ylim(ylim, ylim2)
                if (ymin is not None) and (ymax is not None):
                    ylim  = [ymin, ymax]
                    ylim2 = [ymin, ymax]
                    self.axes.set_ylim(ylim)
                    self.axes2.set_ylim(ylim2)
            # invert y-axis
            ylim = ylim2
            if inv_y2 and (ylim[0] is not None):
                if ylim[0] < ylim[1]:
                    ylim = ylim[::-1]
                self.axes2.set_ylim(ylim)
            else:
                if ylim[1] < ylim[0]:
                    ylim = ylim[::-1]
                self.axes2.set_ylim(ylim)
            # invert x-axis
            inv_x = self.inv_x.get()
            xlim = parse_entry(self.xlim.get())
            if not isinstance(xlim, list):
                xlim = self.axes.get_xlim()
            if inv_x and (xlim[0] is not None):
                if xlim[0] < xlim[1]:
                    xlim = xlim[::-1]
                self.axes.set_xlim(xlim)
            else:
                if xlim[1] < xlim[0]:
                    xlim = xlim[::-1]
                self.axes.set_xlim(xlim)
            # redraw
            self.canvas.draw()
            self.toolbar.update()

    def redraw(self, event=None):
        """
        Redraw the left-hand-side and right-hand-side y-axis.

        Reads the two `y` variable names, the current settings of
        their dimension spinboxes, as well as all other plotting options.
        Then redraws both y-axes.

        """
        # get all states
        # rowxy
        x = self.x.get()
        y = self.y.get()
        # rowy2
        y2 = self.y2.get()

        # Clear both axes first, otherwise x-axis shows only if
        # line2 is chosen.
        self.axes.clear()
        self.axes2.clear()
        self.axes2.yaxis.set_label_position("right")
        self.axes2.yaxis.tick_right()
        # set x, y, axes labels
        if (y != '') or (y2 != ''):
            # y axis
            if y != '':
                vy = vardim2var(y)
                yy = self.df[vy]
                ylab = yy.name
            # y2 axis
            if y2 != '':
                vy2 = vardim2var(y2)
                yy2 = self.df[vy2]
                ylab2 = yy2.name
            if (x != ''):
                # x axis
                vx = vardim2var(x)
                xx = self.df[vx]
                xlab = xx.name
            else:
                # set x to index if not selected
                xx = self.df.index
                xlab = ''
            # set y-axes to nan if not selected
            if (y == ''):
                yy = np.ones_like(xx, dtype='float') * np.nan
                ylab = ''
            if (y2 == ''):
                yy2 = np.ones_like(xx, dtype='float') * np.nan
                ylab2 = ''
            # plot
            # y-axis
            try:
                self.line_y = self.axes.plot(xx, yy)
            except Exception:
                estr = ('Scatter: x (' + vx + ') and y (' + vy + ')'
                        ' shapes do not match for plot:')
                print(estr, xx.shape, yy.shape)
                return
            self.axes.xaxis.set_label_text(xlab)
            self.axes.yaxis.set_label_text(ylab)
            # y2-axis
            try:
                self.line_y2 = self.axes2.plot(xx, yy2)
            except Exception:
                estr  = 'Scatter: x (' + vx + ') and y2 (' + vy2 + ')'
                estr += ' shapes do not match for plot:'
                print(estr, xx.shape, yy2.shape)
                return
            self.axes2.format_coord = lambda x, y: format_coord_scatter(
                x, y, self.axes, self.axes2, xx.dtype, yy.dtype, yy2.dtype)
            self.axes2.xaxis.set_label_text(xlab)
            self.axes2.yaxis.set_label_text(ylab2)
            # styles, invert, same axes, etc.
            self.redraw_y()
            self.redraw_y2()
            # redraw
            self.canvas.draw()
            self.toolbar.update()
        else:
            self.line_y = self.axes.plot([0.], [0.])
            self.canvas.draw()
            self.toolbar.update()



#!/usr/bin/env python
"""
Main dfvue window.

This sets up the main notebook window with the plotting panels.

This module was written by Matthias Cuntz while at Institut National de
Recherche pour l'Agriculture, l'Alimentation et l'Environnement (INRAE), Nancy,
France.

:copyright: Copyright 2023- Matthias Cuntz - mc (at) macu (dot) de
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following classes are provided:

.. autosummary::
   dfvMain

History
   * Written Jul 2023 by Matthias Cuntz (mc (at) macu (dot) de)
   * Use CustomTkinter, Jun 2024, Matthias Cuntz
   * Use mix of grid and pack layout manager, Jun 2024, Matthias Cuntz
   * Use CustomTkinter only if installed, Jun 2024, Matthias Cuntz
   * Back to pack layout manager for resizing, Nov 2024, Matthias Cuntz

"""
import tkinter as tk
import tkinter.ttk as ttk
try:
    from customtkinter import CTkTabview as Frame
    ihavectk = True
except ModuleNotFoundError:
    from tkinter.ttk import Frame
    ihavectk = False
from .dfvscatter import dfvScatter


__all__ = ['dfvMain']


#
# Window with plot panels
#

class dfvMain(Frame):
    """
    Main dfvue notebook window with the plotting panels.

    Sets up the notebook layout with the panels.

    Contains the method to check if csv file has changed.

    """

    #
    # Window setup
    #

    def __init__(self, master, **kwargs):
        super().__init__(master, **kwargs)

        self.name   = 'dfvMain'
        self.master = master      # master window, i.e. root
        self.top    = master.top  # top window

        if ihavectk:
            stab = 'Scatter/Line'
            self.add(stab)
            itab = self.tab(stab)
            itab.name   = self.name
            itab.master = self.master
            itab.top    = self.top
            self.tab_scatter = dfvScatter(itab)
            self.tab_scatter.pack(side=tk.TOP, fill=tk.BOTH, expand=1)
        else:
            # Notebook for tabs for future plot types
            self.tabs = ttk.Notebook(self)
            self.tabs.pack(side=tk.TOP, fill=tk.BOTH, expand=1)
            self.tab_scatter = dfvScatter(self)
            self.tab_scatter.pack(side=tk.TOP, fill=tk.BOTH, expand=1)
            self.tabs.add(self.tab_scatter, text=self.tab_scatter.name)



#!/usr/bin/env python
"""
Scatter/Line panel of dfvue.

The panel allows plotting variables against time or two variables against
each other. A second variable can be plotted in the same graph using the
right-hand-side y-axis.

This module was written by Matthias Cuntz while at Institut National de
Recherche pour l'Agriculture, l'Alimentation et l'Environnement (INRAE), Nancy,
France.

:copyright: Copyright 2023- Matthias Cuntz - mc (at) macu (dot) de
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following classes are provided:

.. autosummary::
   read_csvopts
   read_csvdefaults
   read_csvhelp
   dfvReadcsv

History
   * Written Jul 2023 by Matthias Cuntz (mc (at) macu (dot) de)
   * Use CustomTkinter, Jun 2024, Matthias Cuntz
   * Use mix of grid and pack layout manager, Jun 2024, Matthias Cuntz
   * Use CustomTkinter only if installed, Jun 2024, Matthias Cuntz
   * Concat multiple input files, Oct 2024, Matthias Cuntz
   * Add low_memory to read_csvopts, Jan 2025, Matthias Cuntz
   * Keep na_values as str, Jan 2025, Matthias Cuntz
   * Moved parse_entry to dfvutils, Jan 2025, Matthias Cuntz

"""
import tkinter as tk
import tkinter.ttk as ttk
import tkinter.font as tkfont
try:
    from customtkinter import CTkToplevel as Toplevel
    from customtkinter import CTkFrame as Frame
    from customtkinter import CTkLabel as Label
    from customtkinter import CTkButton as Button
    ihavectk = True
except ModuleNotFoundError:
    from tkinter import Toplevel
    from tkinter.ttk import Frame
    from tkinter.ttk import Label
    from tkinter.ttk import Button
    ihavectk = False
from collections.abc import Iterable
from math import isfinite
import warnings
import pandas as pd
from .dfvutils import parse_entry
from .ncvwidgets import add_entry, add_tooltip, Treeview, callurl


__all__ = ['read_csvopts', 'read_csvdefaults', 'read_csvhelp',
           'dfvReadcsv']


read_csvopts = ['sep', 'index_col', 'usecols', 'skiprows', 'nrows',
                'parse_dates', 'date_format',
                'header', 'names',
                'na_filter', 'na_values', 'keep_default_na',
                'true_values', 'false_values',
                'skip_blank_lines', 'skipinitialspace',
                'comment', 'skipfooter',
                'thousands', 'decimal', 'lineterminator',
                'quotechar', 'doublequote',
                'dayfirst', 'encoding', 'low_memory'
                ]
read_csvopts.append('missing_value')
"""
pandas.read_csv options provided in form in the given order

"""

read_csvdefaults = {'sep': '', 'header': 'infer',
                    'names': '', 'index_col': None, 'usecols': None,
                    'dtype': None, 'engine': None, 'converters': None,
                    'true_values': None, 'false_values': None,
                    'skipinitialspace': False, 'skiprows': None,
                    'skipfooter': 0, 'nrows': None, 'na_values': None,
                    'keep_default_na': True, 'na_filter': True,
                    'verbose': False, 'skip_blank_lines': True,
                    'parse_dates': True,
                    'date_format': None, 'dayfirst': False,
                    'cache_dates': True, 'iterator': False,
                    'chunksize': None, 'compression': 'infer',
                    'thousands': None, 'decimal': '.', 'lineterminator': None,
                    'quotechar': '"', 'quoting': 0, 'doublequote': True,
                    'escapechar': None, 'comment': None, 'encoding': None,
                    'encoding_errors': 'strict', 'dialect': None,
                    'on_bad_lines': 'error', 'delim_whitespace': False,
                    'low_memory': True, 'memory_map': False,
                    'float_precision': None, 'storage_options': None}
read_csvdefaults.update({'missing_value': None})
"""
pandas.read_csv options defaults

"""

read_csvhelp = {
    'file':
    r"""str, path object or file-like object
    Any valid string path is acceptable.
    The string could be a URL. Valid URL schemes include http, ftp, s3, gs,
and file. For file URLs, a host is expected.
    A local file could be: file://localhost/path/to/table.csv.
    If you want to pass in a path object, pandas accepts any os.PathLike.
    By file-like object, we refer to objects with a read() method, such as
a file handle (e.g. via builtin open function) or StringIO.""",
    'sep':
    r"""str, default: ,
    Delimiter to use.
    If sep is None, the C engine cannot automatically detect the separator,
but the Python parsing engine can, meaning the latter will be used and
automatically detect the separator by Python's builtin sniffer tool,
csv.Sniffer.
    In addition, separators longer than 1 character and different from '\s+'
will be interpreted as regular expressions and will also force the use of the
Python parsing engine.
    Note that regex delimiters are prone to ignoring quoted data.
    Regex example: '\r\t'.""",
    'header':
    """int, list of int, None, default: infer
    Row number(s) to use as the column names, and the start of the data.
    Default behavior is to infer the column names: if no names are passed the
behavior is identical to header=0 and column names are inferred from the first
line of the file, if column names are passed explicitly then the behavior is
identical to header=None.
    Explicitly pass header=0 to be able to replace existing names. The header
can be a list of integers that specify row locations for a multi-index on the
columns e.g. 0,1,3. Intervening rows that are not specified will be skipped
(e.g. 2 in this example is skipped).
    Note that this parameter ignores commented lines and empty lines if
skip_blank_lines=True, so header=0 denotes the first line of data rather than
the first line of the file.""",
    'names':
    """array-like, optional
    List of column names to use.
    If the file contains a header row, then you should explicitly pass
header=0 to override the column names.
    Duplicates in this list are not allowed.""",
    'index_col':
    """int, str, sequence of int / str, or False, optional, default: None
    Column(s) to use as the row labels of the DataFrame, either given as
string name (without column index) or column index.
    If a sequence of int / str is given, a MultiIndex is used. dfvue converts
a MultiIndex to a single string index with blanks.
    Note: index_col=False can be used to force pandas to not use the first
column as the index, e.g. when you have a malformed file with delimiters at
the end of each line.""",
    'usecols':
    """list-like or callable, optional
    Return a subset of the columns.
    If list-like, all elements must either be positional (i.e. integer indices
into the document columns) or strings that correspond to column names provided
either by the user in names or inferred from the document header row(s).
    If names are given, the document header row(s) are not taken into account.
For example, a valid list-like usecols parameter would be 0, 1, 2 or
'foo', 'bar', 'baz'.
    Element order is ignored, so usecols=[0, 1] is the same as [1, 0].
    To instantiate a DataFrame from data with element order preserved use
pd.read_csv(data, usecols=['foo', 'bar'])[['foo', 'bar']]
for columns in ['foo', 'bar'] order or
pd.read_csv(data, usecols=['foo', 'bar'])[['bar', 'foo']]
for ['bar', 'foo'] order.
    If callable, the callable function will be evaluated against the column
names, returning names where the callable function evaluates to True. An
example of a valid callable argument would be
lambda x: x.upper() in ['AAA', 'BBB', 'DDD'].
Using this parameter results in much faster parsing time and lower memory
usage.""",
    'dtype':
    """Type name or dict of column -> type, optional
    Data type for data or columns.
    E.g. 'a': np.float64, 'b': np.int32, 'c': 'Int64'
    Use str or object together with suitable na_values settings to preserve and
not interpret dtype.
    If converters are specified, they will be applied INSTEAD of dtype
conversion.""",
    'engine':
    """{'c', 'python', 'pyarrow'}, optional
    Parser engine to use.
    The C and pyarrow engines are faster, while the python engine is currently
more feature-complete. Multithreading is currently only supported by the
pyarrow engine.""",
    'converters':
    """dict, optional
    Dict of functions for converting values in certain columns.
    Keys can either be integers or column labels.""",
    'true_values':
    """list, optional
    Values to consider as True in addition to case-insensitive variants
of 'True'.""",
    'false_values':
    """list, optional
    Values to consider as False in addition to case-insensitive variants
of 'False'.""",
    'skipinitialspace':
    """bool, default: False
    Skip spaces after delimiter.""",
    'skiprows':
    """list-like, int or callable, optional
    Line numbers to skip (0-indexed, must include comma, e.g. 1, for skipping
the second row) or number of lines to skip (int, without comma) at the start of
the file.
    If callable, the callable function will be evaluated against the row
indices, returning True if the row should be skipped and False otherwise.
An example of a valid callable argument would be lambda x: x in [0, 2].""",
    'skipfooter':
    """int, default: 0
    Number of lines at bottom of file to skip
(Unsupported with engine='c').""",
    'nrows':
    """int, optional
    Number of rows of file to read.
    Useful for reading pieces of large files.""",
    'na_values':
    """scalar, str, list-like, or dict, optional
    Additional strings to recognize as NA/NaN.
    If dict passed, specific per-column NA values.
    By default the following values are interpreted as NaN:
'', '#N/A', '#N/A N/A', '#NA', '-1.#IND', '-1.#QNAN', '-NaN', '-nan',
'1.#IND', '1.#QNAN', '<NA>', 'N/A', 'NA', 'NULL', 'NaN', 'None', 'n/a',
'nan', 'null'.""",
    'keep_default_na':
    """bool, default: True
    Whether or not to include the default NaN values when parsing the data.
    Depending on whether na_values is passed in, the behavior is as follows:
        If keep_default_na is True, and na_values are specified,
na_values is appended to the default NaN values used for parsing.
        If keep_default_na is True, and na_values are not specified,
only the default NaN values are used for parsing.
        If keep_default_na is False, and na_values are specified,
only the NaN values specified na_values are used for parsing.
        If keep_default_na is False, and na_values are not specified,
no strings will be parsed as NaN.
    Note that if na_filter is passed in as False, the keep_default_na and
na_values parameters will be ignored.""",
    'na_filter':
    """bool, default: True
    Detect missing value markers (empty strings and the value of na_values).
    In data without any NAs, passing na_filter=False can improve the
performance of reading a large file.""",
    'verbose':
    """bool, default: False
    Indicate number of NA values placed in non-numeric columns.""",
    'skip_blank_lines':
    """bool, default: True
    If True, skip over blank lines rather than interpreting as NaN values.""",
    'parse_dates':
    """bool or list of int or names or list of lists or dict, default: False
    The behavior is as follows:
        boolean. If True -> try parsing the index.
        list of int or names. e.g. If 1, 2, 3
-> try parsing columns 1, 2, 3 each as a separate date column.
        list of lists. e.g. If [1, 3]
-> combine columns 1 and 3 and parse as a single date column.
        dict, e.g. 'foo' : [1, 3]
-> parse columns 1, 3 as date and call result 'foo'
    If a column or index cannot be represented as an array of datetimes,
say because of an unparsable value or a mixture of timezones, the column or
index will be returned unaltered as an object data type. For non-standard
datetime parsing, use pd.to_datetime after pd.read_csv.
    Note: A fast-path exists for iso8601-formatted dates.""",
    'date_format':
    """str or dict of column -> format, default: None
    If used in conjunction with parse_dates, will parse dates according to
this format.
    For anything more complex, please read in as object and then apply
to_datetime() as-needed.""",
    'dayfirst':
    """bool, default: False
    DD/MM format dates, international and European format.""",
    'cache_dates':
    """bool, default: True
    If True, use a cache of unique, converted dates to apply the datetime
conversion.
    May produce significant speed-up when parsing duplicate date strings,
especially ones with timezone offsets.""",
    'iterator':
    """bool, default: False
    Return TextFileReader object for iteration or getting chunks
with get_chunk().""",
    'chunksize':
    """int, optional
    Return TextFileReader object for iteration.
    See the IO Tools docs for more information on iterator and chunksize.""",
    'compression':
    """str or dict, default: 'infer'
    For on-the-fly decompression of on-disk data.
    If 'infer' and 'filepath_or_buffer' is path-like, then detect compression
from the following extensions:
'.gz', '.bz2', '.zip', '.xz', '.zst', '.tar', '.tar.gz', '.tar.xz' or
'.tar.bz2' (otherwise no compression).
    If using 'zip' or 'tar', the ZIP file must contain only one data file to
be read in.
    Set to None for no decompression.
    Can also be a dict with key 'method' set to one of
{'zip', 'gzip', 'bz2', 'zstd', 'tar'} and other key-value pairs are forwarded
to zipfile.ZipFile, gzip.GzipFile, bz2.BZ2File, zstandard.ZstdDecompressor or
tarfile.TarFile, respectively. As an example, the following could be passed for
Zstandard decompression using a custom compression dictionary:
compression={'method': 'zstd', 'dict_data': my_compression_dict}.""",
    'thousands':
    """str, optional
    Thousands separator.""",
    'decimal':
    """str, default: '.'
    Character to recognize as decimal point
(e.g. use ',' for European data).""",
    'lineterminator':
    """str (length 1), optional
    Character to break file into lines. Only valid with C parser.""",
    'quotechar':
    """str (length 1), optional
    The character used to denote the start and end of a quoted item.
    Quoted items can include the delimiter and it will be ignored.""",
    'quoting':
    """int or csv.QUOTE_* instance, default: 0
    Control field quoting behavior per csv.QUOTE_* constants.
    Use one of QUOTE_MINIMAL (0), QUOTE_ALL (1), QUOTE_NONNUMERIC (2) or
QUOTE_NONE (3).""",
    'doublequote':
    """bool, default: True
    When quotechar is specified and quoting is not QUOTE_NONE, indicate whether
or not to interpret two consecutive quotechar elements INSIDE a field as a
single quotechar element.""",
    'escapechar':
    """str (length 1), optional
    One-character string used to escape other characters.""",
    'comment':
    r"""str, optional
    Indicates remainder of line should not be parsed.
    If found at the beginning of a line, the line will be ignored altogether.
    This parameter must be a single character. Like empty lines
(as long as skip_blank_lines=True), fully commented lines are ignored by the
parameter header but not by skiprows.
    For example, if comment='#', parsing
        #empty\na,b,c\n1,2,3
    with header=0 will result in 'a,b,c' being treated as the header.""",
    'encoding':
    """str, optional, default: 'utf-8'
    Encoding to use for UTF when reading/writing (ex. 'utf-8').
    List of Python standard encodings.""",
    'encoding_errors':
    """str, optional, default: 'strict'
    How encoding errors are treated.
    List of possible values.""",
    'dialect':
    """str or csv.Dialect, optional
    If provided, this parameter will override values (default or not) for the
following parameters: delimiter, doublequote, escapechar, skipinitialspace,
quotechar, and quoting.
    If it is necessary to override values, a ParserWarning will be issued.
See csv.Dialect documentation for more details.""",
    'on_bad_lines':
    """{'error', 'warn', 'skip'} or callable, default: 'error'
    Specifies what to do upon encountering a bad line
(a line with too many fields).
    Allowed values are:
        'error', raise an Exception when a bad line is encountered.
        'warn', raise a warning when a bad line is encountered and skip
that line.
        'skip', skip bad lines without raising or warning when they are
encountered.""",
    'delim_whitespace':
    r"""bool, default: False
    Specifies whether or not whitespace (e.g. ' ' or ' ') will be used
as the sep.
    Equivalent to setting sep='\s+'.
    If this option is set to True, nothing should be passed in for the
delimiter parameter.""",
    'low_memory':
    """bool, default: True
    Internally process the file in chunks, resulting in lower memory use
while parsing, but possibly mixed type inference.
    To ensure no mixed types either set False, or specify the type with
the dtype parameter.
    Note that the entire file is read into a single DataFrame regardless,
use the chunksize or iterator parameter to return the data in chunks.
(Only valid with C parser).""",
    'memory_map':
    """bool, default: False
    If a filepath is provided for filepath_or_buffer, map the file object
directly onto memory and access the data directly from there.
    Using this option can improve performance because there is no longer
any I/O overhead.""",
    'float_precision':
    """str, optional
    Specifies which converter the C engine should use for
floating-point values.
    The options are
        None or 'high' for the ordinary converter,
        'legacy' for the original lower precision pandas converter, and
        'round_trip' for the round-trip converter.""",
    'storage_options':
    """dict, optional
    Extra options that make sense for a particular storage connection,
e.g. host, port, username, password, etc.
    For HTTP(S) URLs the key-value pairs are forwarded to
urllib.request.Request as header options.
    For other URLs (e.g. starting with 's3://', and 'gcs://') the key-value
pairs are forwarded to fsspec.open.
    Please see fsspec and urllib for more details, and for more examples
on storage options refer here.""",
    'dtype_backend':
    """{'numpy_nullable', 'pyarrow'}, defaults to NumPy backed DataFrames
    Which dtype_backend to use, e.g. whether a DataFrame should have
NumPy arrays, nullable dtypes are used for all dtypes that have a nullable
implementation when 'numpy_nullable' is set, pyarrow is used for all dtypes
if 'pyarrow' is set.
    The dtype_backends are still experimental.""",
}
read_csvhelp.update({'missing_value':
                     'scalar or str\n'
                     '    Additional value to na_values set to NA/NaN.'})
"""
Tooltips for pandas.read_csv options from pandas v2.0.3

"""


class dfvReadcsv(Toplevel):
    """
    Window for reading csv files.

    """

    #
    # Setup panel
    #

    def __init__(self, top, callback=None, **kwargs):
        super().__init__(top, **kwargs)

        self.top = top  # top window
        self.callback = callback

        self.name = 'dfvReadcsv'
        self.title("Read csv file")
        if ihavectk:
            self.geometry('1000x540+100+10')
        else:
            self.geometry('1000x490+100+10')

        # copy for ease of use
        self.csvfile = self.top.csvfile
        self.newcsvfile = self.top.newcsvfile
        self.df = self.top.df
        self.sep = self.top.sep
        self.index_col = self.top.index_col
        self.skiprows = self.top.skiprows
        self.parse_dates = self.top.parse_dates
        self.date_format = self.top.date_format
        self.missing_value = self.top.missing_value
        self.cols = self.top.cols

        # space
        rootspace = Frame(self)
        rootspace.pack(side='top', fill='x')
        rootspacespace = Label(rootspace, text=" ")
        rootspacespace.pack(side='left')

        # 1. row - treeview current DataFrame
        self.rowtree = Frame(self)
        self.rowtree.pack(side='top', fill='x')
        self.nrows = 10
        if self.df is None:
            print('No df')
            ncols = 5
            columns = [ f'Column {i:03d}' for i in range(ncols) ]
            dat = [ [''] * ncols for i in range(self.nrows * 4) ]
            df = pd.DataFrame(dat, columns=columns)
        else:
            df = self.df
        self.new_tree(df)

        # space
        treespace = Frame(self)
        treespace.pack(side='top', fill='x')
        treespacespace = Label(treespace, text=" ")
        treespacespace.pack(side='left')

        # label for read_csv options
        opthead = Frame(self)
        opthead.pack(side='top', fill='x')
        optheadlabel1 = Label(opthead, text='Options for ')
        optheadlabel1.pack(side='left')
        if ihavectk:
            optheadlabel2 = Label(opthead, text='pandas.read_csv',
                                  text_color=('blue', 'lightblue'))
        else:
            # https://stackoverflow.com/questions/1529847/how-to-change-the-foreground-or-background-colour-of-a-tkinter-button-on-mac-os/42591118#42591118
            ttk.Style().configure('blue.TLabel', foreground='#0096FF')
            optheadlabel2 = Label(opthead, text='pandas.read_csv',
                                  style='blue.TLabel')
            # https://stackoverflow.com/questions/3655449/underline-text-in-tkinter-label-widget
            font = tkfont.Font(optheadlabel2, optheadlabel2.cget("font"))
            font.configure(underline=True)
            optheadlabel2.configure(font=font)
        optheadlabel2.pack(side='left')
        optheadlabel2.bind("<Button-1>",
                           lambda e:
                           callurl("https://pandas.pydata.org/docs/reference/"
                                   "api/pandas.read_csv.html"))
        optheadlabel3 = Label(opthead, text=' (date_format: see ')
        optheadlabel3.pack(side='left')
        if ihavectk:
            optheadlabel4 = Label(opthead, text='strftime',
                                  text_color=('blue', 'lightblue'))
        else:
            optheadlabel4 = Label(opthead, text='strftime',
                                  style='blue.TLabel')
            optheadlabel4.configure(font=font)
        optheadlabel4.pack(side='left')
        optheadlabel4.bind("<Button-1>",
                           lambda e:
                           callurl("https://docs.python.org/3/library/"
                                   "datetime.html#"
                                   "strftime-and-strptime-behavior"))
        optheadlabel5 = Label(opthead, text=')')
        optheadlabel5.pack(side='left')

        # option fields
        self.optframe = {}  # entry frame
        self.optlbl = {}    # label
        self.opt = {}       # entry test
        self.opttip = {}    # tooltip

        # rows with pandas.read_csv options
        oend = 0
        nopt = len(read_csvopts)
        noptrow = nopt // 4
        if (nopt % 5) > 0:
            noptrow += 1
        if ihavectk:
            entrywidth = 85  # px
            padlabel = 3     # characters
            padxlabel = 5    # px
        else:
            entrywidth = 8   # characters
            padlabel = 5     # characters
            padxlabel = 5    # px
        self.rowopt = Frame(self)
        self.rowopt.pack(side='top', fill='x')
        for nr in range(noptrow):
            ostart = oend
            oend = ostart + 5
            for oois, oo in enumerate(read_csvopts[ostart:oend]):
                (self.optframe[oo], self.optlbl[oo],
                 self.opt[oo], self.opttip[oo]) = add_entry(
                     self.rowopt, label=oo,
                     padlabel=padlabel,
                     text=read_csvdefaults[oo],
                     width=entrywidth, padx=padxlabel,
                     tooltip=read_csvhelp[oo],
                     command=[self.read_again, self.read_final])
                self.optframe[oo].grid(row=nr, column=2 * oois, columnspan=2,
                                       sticky=tk.E)
                # self.optframe[oo].pack(side='right')
        # overwrite defaults from command line
        self.set_command_line_options(defaults=self.newcsvfile)

        # add cancel and read buttons to last row
        self.rowdone = Frame(self)
        self.rowdone.pack(side='top', fill='x')
        self.done = Button(self.rowdone, text="Read csv",
                           command=self.read_final)
        self.donetip = add_tooltip(self.done, 'Finished reading csv')
        self.done.pack(side='right', padx=10, pady=5)

        self.cancel = Button(self.rowdone, text="Cancel",
                             command=self.cancel)
        self.canceltip = add_tooltip(self.cancel, 'Cancel reading csv file')
        self.cancel.pack(side='right', pady=5)

        self.focus_force()

        self.read_again('')

    #
    # Event bindings
    #

    def cancel(self, event=None):
        if self.callback is not None:
            self.callback()
        # do not self.destroy() with ctk.CTkButton, leading to
        # 'invalid command name
        #     ".!dfvreadcsv.!ctkframe3.!ctkbutton2.!ctkcanvas"'
        # self.destroy() works with ttk.Button
        self.withdraw()

    def new_tree(self, df):
        """
        Make new Treeview widget and fill it with pandas.Dataframe

        """
        # create
        self.tree = Treeview(self.rowtree, xscroll=True, yscroll=True)
        self.tree.pack()
        self.tree.tag_configure("even", background='white',
                                foreground='black')
        self.tree.tag_configure("odd", background='gray',
                                foreground='white')
        # fill
        idx = 'index'
        if df.index.name is not None:
            idx = 'index ' + df.index.name
        columns = [idx]
        df_columns = []  # index: name
        for ii, cc in enumerate(list(df.columns)):
            df_columns.append(f'{ii}: {cc}')
        columns.extend(df_columns)  # index, df_columns
        self.tree.config(columns=columns, show="headings",
                         height=self.nrows)
        # columns
        self.tree.column(idx, width=150, stretch=False,
                         anchor='center')
        if len(df.columns) == 1:
            # if sep not correct show long line
            cwidth = 700
        else:
            cwidth = 150
        for c in df_columns:  # columns w/o index
            self.tree.column(c, width=cwidth, stretch=False,
                             anchor='center')
        for c in columns:  # columns with index
            self.tree.heading(c, text=c, anchor='center')
        # rows
        for i in range(min([self.nrows * 4, df.shape[0]])):
            values = [df.index[i]]
            values.extend(list(df.iloc[i].values))
            self.tree.insert(
                '', 'end', values=values,
                tags=('even',) if (i % 2) == 0 else ('odd',) )

    def read_again(self, event):
        self.tree.destroy()
        self.read_df(nrows=4 * self.nrows)
        self.new_tree(self.df)

    def read_df(self, nrows=None):
        """
        pandas.read_csv using options from entry fields

        """
        opts = {}
        for oo in read_csvopts:
            text = self.opt[oo].get()
            tt = parse_entry(text)
            if (tt != '') and (tt is not None):
                opts.update({oo: tt})
        # skipfooter does not work with nrows
        if nrows is not None:
            opts['nrows'] = nrows
            del opts['skipfooter']
        # add missing_value to na_values
        if 'missing_value' in opts:
            if 'na_values' in opts:
                if isinstance(opts['na_values'], Iterable):
                    nn = list(opts['na_values'])
                    nn.append(opts['missing_value'])
                    opts['na_values'] = nn
                else:
                    opts['na_values'] = [opts['na_values'],
                                         opts['missing_value']]
            else:
                opts['na_values'] = opts['missing_value']
            del opts['missing_value']
        # date_format introduced in pandas v2.0.0
        if pd.__version__ < '2':
            if 'date_format' in opts:
                date_parser = lambda date: pd.to_datetime(
                    date, format=opts['date_format'])
                opts['date_parser'] = date_parser
                del opts['date_format']
        # iparsedates = False
        # if pd.__version__ > '2.0':
        #     if 'parse_dates' in opts:
        #         if not all(o.__hash__ is not None
        #                    for o in opts['parse_dates']):
        #             # parse_date is list of int, list of list, or dict
        #             iparsedates = True
        #             parse_dates = opts['parse_dates']
        #             # del opts['parse_dates']
        #             if 'date_format' in opts:
        #                 date_format = opts['date_format']
        #                 # del opts['date_format']
        #             if 'index_col' in opts:
        #                 index_col = opts['index_col']
        #                 # del opts['index_col']
        # # Testing
        # try:
        #     self.df = pd.read_csv(self.csvfile, **opts)
        # except TypeError:
        #     print('Did not work')
        #     pass
        with warnings.catch_warnings():
            warnings.simplefilter(action='ignore', category=FutureWarning)
            if (nrows is None) and (len(self.csvfile) > 1):
                dfl = []
                for cfile in self.csvfile:
                    dfl.append(pd.read_csv(cfile, **opts))
                self.df = pd.concat(dfl)
            else:
                self.df = pd.read_csv(self.csvfile[0], **opts)
        # if iparsedates:
        #     True
        # # Transformation
        # if nrows is None:
        #     self.df = self.df.resample('1D').mean().squeeze()

    def read_final(self, event=None):
        self.tree.destroy()
        self.read_df()
        # add index as column
        idx = 'df.index'
        if self.df.index.name is not None:
            idx = self.df.index.name
        if isinstance(self.df.index, pd.MultiIndex):
            series = [ [ str(kk) for kk in ii ]
                       for ii in self.df.index ]
            series = [ ' '.join(ii) for ii in series ]
        else:
            series = self.df.index
        self.df.insert(0, idx, pd.Series(series, index=self.df.index))
        if isinstance(self.df.index, pd.MultiIndex):
            # replace MultiIndex with combined column
            self.df.set_index(idx, drop=False, inplace=True)
        rows = self.df.shape[0]
        self.cols = [ f'{cc} ({rows} {self.df[cc].dtype.name})'
                      for cc in self.df.columns ]
        self.top.df = self.df
        self.top.cols = self.cols
        if self.callback is not None:
            self.callback()
        # do not self.destroy() with ctk.CTkButton, leading to
        # 'invalid command name
        #     ".!dfvreadcsv.!ctkframe3.!ctkbutton2.!ctkcanvas"'
        # self.destroy() works with ttk.Button
        self.withdraw()

    def set_command_line_options(self, defaults=False):
        """
        Set options possible on the command line

        If defaults=True, use default options otherwise set options
        from command line.

        """
        if defaults:
            self.opt['sep'].set('')
            self.opt['index_col'].set(None)
            self.opt['skiprows'].set(None)
            self.opt['parse_dates'].set('True')
            self.opt['date_format'].set(None)
            self.opt['missing_value'].set(None)
        else:
            if self.sep != '':
                self.opt['sep'].set(self.sep)
            if self.index_col is not None:
                self.opt['index_col'].set(self.index_col)
            if self.skiprows is not None:
                self.opt['skiprows'].set(self.skiprows)
            if self.parse_dates is not None:
                self.opt['parse_dates'].set(self.parse_dates)
            if self.date_format is not None:
                self.opt['date_format'].set(self.date_format)
            if self.missing_value is not None:
                self.opt['missing_value'].set(self.missing_value)
        return



#!/usr/bin/env python
"""
usage: dfvue [-h] [-m missing_value] [csv_file]

A minimal GUI for a quick view of csv files.

positional arguments:
  csv_file              Delimited text file

optional arguments:
  -h, --help            show this help message and exit
  -m missing_value, --miss missing_value
                        Missing or undefined value set to NaN.
                        For negative values, use for example --miss=-9999.


Example command line:
    dfvue meteo_DB1_2020.csv

:copyright: Copyright 2023- Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

History
   * Written Jul 2023 2023 by Matthias Cuntz (mc (at) macu (dot) de)
   * Allow multiple input files, Oct 2024, Matthias Cuntz
   * Default [''] for csvfile instead of None, Jan 2025, Matthias Cuntz

"""
from dfvue import dfvue


def main():
    import argparse

    sep = ''
    index_col = None
    skiprows = None
    parse_dates = 'True'
    date_format = None
    missing_value = None
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description='A minimal GUI for a quick view of csv files.')
    hstr = 'Delimiter to use.'
    parser.add_argument('-s', '--sep', action='store', type=str,
                        default=sep, dest='sep',
                        metavar='separator', help=hstr)
    hstr = ('Column(s) to use as index, either given as column index'
            ' or string name.')
    parser.add_argument('-i', '--index_col', action='store', type=str,
                        default=index_col, dest='index_col',
                        metavar='columns', help=hstr)
    hstr = ('Line number(s) to skip (0-indexed, must include comma,'
            ' e.g. "1," for skipping the second row) or\n'
            'number of lines to skip (int, without comma) at the start'
            ' of the file.')
    parser.add_argument('-k', '--skiprows', action='store', type=str,
                        default=skiprows, dest='skiprows',
                        metavar='rows', help=hstr)
    hstr = ('boolean, if True -> try parsing the index.\n'
            'list of int or names, e.g. 1,2,3\n'
            '    -> try parsing columns 1, 2, and 3 each as a separate date'
            ' column.\n'
            'list of lists, e.g. [1,3]\n'
            '    -> combine columns 1 and 3 and parse as a single date'
            ' column.\n'
            'dict, e.g. "foo":[1,3]\n'
            '    -> parse columns 1 and 3 as date and call result "foo"')
    parser.add_argument('-p', '--parse_dates', action='store', type=str,
                        default=parse_dates, dest='parse_dates',
                        metavar='bool/list/dict', help=hstr)
    hstr = ('Will parse dates according to this format.\n'
            'For example: "%%Y-%%m-%%d %%H:%%M%%S". See\n'
            'https://docs.python.org/3/library/datetime.html#strftime-and-strptime-behavior')
    parser.add_argument('-d', '--date_format', action='store', type=str,
                        default=date_format, dest='date_format',
                        metavar='format_string', help=hstr)
    hstr = ('Missing or undefined value set to NaN.\n'
            'For negative values, use long format, e.g. --missing_value=-9999.')
    parser.add_argument('-m', '--missing_value', action='store', type=str,
                        default=missing_value, dest='missing_value',
                        metavar='missing_value', help=hstr)
    parser.add_argument('csvfile', nargs='*', default=[''],
                        metavar='csv_file(s)',
                        help='Delimited text file(s)')

    args = parser.parse_args()
    sep = args.sep
    index_col = args.index_col
    skiprows = args.skiprows
    parse_dates = args.parse_dates
    date_format = args.date_format
    missing_value = args.missing_value
    csvfile = args.csvfile

    del parser, args

    # This must be before any other call to matplotlib
    # because it uses the TkAgg backend.
    # This means, do not use --pylab with ipython.
    dfvue(csvfile=csvfile, sep=sep, index_col=index_col, skiprows=skiprows,
          parse_dates=parse_dates, date_format=date_format,
          missing_value=missing_value)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
Purpose
-------

dfvue is a minimal GUI for a quick view of csv files.

:copyright: Copyright 2023- Matthias Cuntz - mc (at) macu (dot) de
:license: MIT License, see LICENSE for details.

Subpackages
-----------
.. autosummary::
   dfvmain
   dfvreadcsv
   dfvscatter
   dfvtransform
   dfvue
   dfvutils
   ncvwidgets

History
   * Written Jul 2023 by Matthias Cuntz (mc (at) macu (dot) de)

"""
# version, author
try:
    from ._version import __version__
except ImportError:  # pragma: nocover
    # package is not installed
    __version__ = "0.0.0.dev0"
__author__  = "Matthias Cuntz"

# helper
# copy from idle
from .tooltip import TooltipBase, OnHoverTooltipBase, Hovertip

# general helper functions
from .dfvutils import clone_dfvmain, format_coord_scatter, list_intersection
from .dfvutils import parse_entry, vardim2var

# adding widgets with labels, etc.
from .ncvwidgets import callurl, Tooltip
from .ncvwidgets import add_checkbutton, add_combobox, add_entry, add_imagemenu
from .ncvwidgets import add_menu, add_scale, add_spinbox, add_tooltip
from .ncvwidgets import Treeview

# panels
# read csv window
from .dfvreadcsv import read_csvopts, read_csvdefaults, read_csvhelp
from .dfvreadcsv import dfvReadcsv

# scatter/line panel
from .dfvscatter import dfvScatter

# manipulate DataFrame
from .dfvtransform import dfvTransform

# main window with panels
from .dfvmain import dfvMain

# main
from .dfvue import dfvue


__all__ = ['TooltipBase', 'OnHoverTooltipBase', 'Hovertip',
           'clone_dfvmain',
           'format_coord_scatter',
           'list_intersection', 'parse_entry', 'vardim2var',
           'Tooltip',
           'add_checkbutton', 'add_combobox', 'add_entry', 'add_imagemenu',
           'add_menu', 'add_scale', 'add_spinbox', 'add_tooltip',
           'Treeview',
           'read_csvopts', 'read_csvdefaults', 'read_csvhelp',
           'dfvReadcsv',
           'dfvScatter',
           'dfvTransform',
           'dfvMain',
           'dfvue',
           ]
